module Search where

import Char
import Swords

import Control.Monad
import Data.List
import Data.Maybe
import System.Random
import qualified Data.Set as S

type Gen = IO

rnd :: Int -> Gen Int
rnd n = randomRIO (0, n-1)

genChrom :: Gen Chrom
genChrom = improve =<< replicateM 10 genGrowth

genChromG :: Swordset -> Gen Chrom
genChromG ss = improveG ss =<< replicateM 10 (genGrowthBySword $ genSwordG ss)

genIndv :: Gen Indv
genIndv = liftM indv genChrom

genChar :: Gen Character
genChar = liftM chromToChar $ replicateM 10 genGrowth

data Growth = GHP | GAtk | GDef | GSpd | GSword Sword
  deriving (Eq, Ord)
type Chrom = [Growth]
type Indv = (Chrom, Character)

instance Show Growth where
  show g = case g of
    GHP -> "HP"
    GAtk -> "攻撃力"
    GDef -> "防御力"
    GSpd -> "素早さ"
    GSword s -> show s

chromToChar :: Chrom -> Character
chromToChar = foldr applyGrowth initialChar

indv :: Chrom -> Indv
indv ch = (ch, chromToChar ch)

improve :: Chrom -> Gen Chrom
improve = improveG m100SS

improveG :: Swordset -> Chrom -> Gen Chrom
improveG ss c = return c >>= addMPGainer ss >>= avoidZeroHp ss >>= avoidZeroAtk ss -- >>= arrangeRitual

avoidZeroHp :: Swordset -> Chrom -> Gen Chrom
avoidZeroHp world ch
  | not $ any (\c -> c == GHP || maybe False swRapid (unGSword c)) ch
    = do
        k <- choose $ GHP : map GSword (ssProspectlike world)
        chooseReplace k ch
  | otherwise = return ch
  where
    swRapid = flip elem $ ssRapid world

avoidZeroAtk :: Swordset -> Chrom -> Gen Chrom
avoidZeroAtk world ch
  | all (/=GAtk) ch && all (flip notElem $ map GSword $ ssHeatlike world) ch && any (maybe False need_atk . unGSword) ch
    = do
        k <- choose [GAtk, GSword heat]
        chooseReplace k ch
  | otherwise = return ch
  where
    need_atk sw = isAtk sw && not (S.member sw $ ssAtkIndependentAttacks world)

addMPGainer :: Swordset -> Chrom -> Gen Chrom
addMPGainer world ch = case (ssLimitedMPGainer world, ssUnlimitedMPGainer world) of
  (Just (lm, ls), Just (um, us)) -> addMPGainer_ lm ls um us ch
  _ -> return ch

addMPGainer_ lm ls um us ch = foldM (flip chooseReplace) ch $ map GSword $ replicate p_num prayer ++ replicate c_num us
  where
    ss = mapMaybe unGSword ch
    (p_need, c_need) = foldl pp (0, 0) $ map need ss
    p_there = count ls
    c_there = count us
    count x = fromIntegral $ length $ filter (==x) ss
    pp (a, b) (c, d) = (a+c, b+d)
    need sw
      | swLimited sw = (mp sw / fromIntegral lm, 0)
      | otherwise = (0, mp sw / fromIntegral um)
    mp = fromIntegral . fromMaybe 0 . swMP
    p_num = min 4 $ ceiling $ p_need - p_there
    c_num = min 2 $ ceiling $ c_need - c_there

arrangeRitual :: Chrom -> Gen Chrom
arrangeRitual ch = return $ replaceSwords ch $ rs ++ nrs
  where
    (rs, nrs) = partition is_ritual ss
    ss = mapMaybe unGSword ch
    is_ritual = (=='儀') . head . swShortName

replaceSwords :: Chrom -> [Sword] -> Chrom
replaceSwords = go
  where
    go [] _ = []
    go ch [] = ch
    go (GSword _:ch) (s:r) = GSword s : go ch r
    go (g:ch) ss = g : go ch ss

unGSword :: Growth -> Maybe Sword
unGSword g = case g of
  GSword s -> Just s
  _ -> Nothing

genGrowthBySword :: Gen Sword -> Gen Growth
genGrowthBySword gsw = chooseM
  [ choose [GHP, GAtk, GDef]
  , return GSpd
  , liftM GSword gsw
  ]

genGrowth :: Gen Growth
genGrowth = genGrowthBySword genSword

genSword :: Gen Sword
genSword = genSwordG m100SS

genSwordG :: Swordset -> Gen Sword
genSwordG ss = chooseM $ map return (ssConstantSwords ss) ++ map (param 1) (ssVariableSwordsPositive ss) ++ map (param 0) (ssVariableSwordsNatural ss) 
  ++ map return (concatMap (replicate 5) $ ssVeryImportantSwords ss)
  where
    param begin gen = do
      n <- rnd 10
      p <- choose $ if n == 0 then [2..20] else [begin..1]
      return $ gen p

data Swordset = Swordset
  { ssConstantSwords :: [Sword]
  , ssVariableSwordsPositive :: [Int -> Sword]
  , ssVariableSwordsNatural :: [Int -> Sword]
  , ssVeryImportantSwords :: [Sword]
  , ssUnlimitedMPGainer :: Maybe (Int, Sword)
  , ssLimitedMPGainer :: Maybe (Int, Sword)
  , ssHeatlike :: [Sword] -- swords that give atk to zero-atk kenshi
  , ssProspectlike :: [Sword] -- swords that give hp to zero-hp kenshi
  , ssAtkIndependentAttacks :: S.Set Sword
  , ssRapid :: [Sword]
  -- ...
  }

emptySS :: Swordset
emptySS = Swordset
  { ssConstantSwords = []
  , ssVariableSwordsPositive = []
  , ssVariableSwordsNatural = []
  , ssVeryImportantSwords = []
  , ssUnlimitedMPGainer = Nothing
  , ssLimitedMPGainer = Nothing
  , ssHeatlike = []
  , ssProspectlike = []
  , ssAtkIndependentAttacks = S.empty
  , ssRapid = []
  }

m100SS :: Swordset
m100SS = Swordset
  { ssConstantSwords = constantSwords
  , ssVariableSwordsPositive = variableSwordsPositive
  , ssVariableSwordsNatural = variableSwordsNatural
  , ssVeryImportantSwords = [derive 0]
  , ssUnlimitedMPGainer = Just (1, chant)
  , ssLimitedMPGainer = Just (3, prayer)
  , ssHeatlike = [heat]
  , ssProspectlike = [prospect]
  , ssAtkIndependentAttacks = S.fromList $
      [magical, death, nuru, spark, firecracker, eruption, devour, disaster, perish, finale]
      ++ map rage [1..20]
      ++ [cool, decay, decel, gentle, flashfire]
  , ssRapid = [rapid, sonic, light, derive 0, prospect]
  }

basicSS :: Swordset
basicSS = emptySS
  { ssConstantSwords = [slash, rapid, heavy, shock, poison, absolute, 
      magical, death, heat, protect, accel, shield, life, mirror]
  , ssVariableSwordsPositive = [recovery]
  , ssHeatlike = [heat]
  , ssAtkIndependentAttacks = S.fromList [magical, death]
  , ssRapid = [rapid]
  }

m2011aSS :: Swordset
m2011aSS = emptySS
  { ssConstantSwords = ssConstantSwords basicSS ++
      [ nuru, spark, perishA, rageA, sonic, light, solid, fang
      , two, three, four, five, six, seven, dance, scatter, toxic
      , abortionA, coolA, decayA, decelA, destroy, battle, confusion
      , bloodsuck, robbery, darkA, antimagicA, soul, pushA, attractionA
      , first, second, healA, chantA, rogation, ritualA, prospect
      , blaze, limit, deriveA, slow, danger, vibration, toxicant, diffuse
      , recurrenceA, resonanceA, fictional, replica, multi, ageA
      , mkbubble, mkfence, mkgate, mkice, mkmist, mktear
      , recoveryA
      , prayer
      ]
  , ssHeatlike = [heat]
  , ssProspectlike = [prospect]
  , ssUnlimitedMPGainer = Just (1, chantA)
  , ssLimitedMPGainer = Just (3, prayer)
  , ssAtkIndependentAttacks = S.fromList [magical, death, nuru, spark, perishA, rageA]
  , ssRapid = [rapid, sonic, light, deriveA, prospect, chantA, rogation]
  }

m2011SS :: Swordset
m2011SS = m2011aSS
  { ssConstantSwords = m2011SpecificSwords ++ ssConstantSwords m2011aSS
  }

mvoidSS :: Swordset
mvoidSS = emptySS
  { ssConstantSwords =
      [ slash, darkA, shock, poison, toxic, battle, confusion, robbery
      , combo, solid, absolute, heavyV, magical, death, rageA, perishA
      , heat, protect, accel, blaze, shieldV, life, soul, mirror, pushA
      , first, recoveryA, healV, ritualA, prospect, deriveA, vibration
      , toxicant, vigorous, ageA
      ]
  , ssHeatlike = [heat]
  , ssProspectlike = [prospect]
  , ssAtkIndependentAttacks = S.fromList [magical, death, rageA, perishA]
  , ssRapid = [prospect, deriveA]
  }

ssContent :: Swordset -> [Sword]
ssContent ss = ssConstantSwords ss
  ++ liftM2 ($) (ssVariableSwordsPositive ss) [1..20]
  ++ liftM2 ($) (ssVariableSwordsNatural ss) [0..20]

applyGrowth :: Growth -> Character -> Character
applyGrowth g ch = case g of
  GHP -> ch{ chHp = chHp ch + 5 }
  GAtk -> ch{ chAtk = chAtk ch + 1 }
  GDef -> ch{ chDef = chDef ch + 1 }
  GSpd -> ch{ chSpd = chSpd ch + 1 }
  GSword s -> ch{ chSwords = s : chSwords ch }

applyGrowthG :: Growth -> Character -> Gen Character
applyGrowthG g ch = case g of
  GHP -> return ch{ chHp = chHp ch + 5 }
  GAtk -> return ch{ chAtk = chAtk ch + 1 }
  GDef -> return ch{ chDef = chDef ch + 1 }
  GSpd -> return ch{ chSpd = chSpd ch + 1 }
  GSword s -> do
    (a, b) <- chooseSplit $ chSwords ch
    return ch{ chSwords = a ++ s : b }

weaken :: Character -> Gen (Growth, Character)
weaken ch@Ch{ chHp = hp, chAtk = atk, chDef = def, chSpd = spd, chSwords = swords } = join $ chooseRated
  [ (fi hp, return (GHP, ch{ chHp = hp - 5 }))
  , (fi atk, return (GAtk, ch{ chAtk = atk - 1 }))
  , (fi def, return (GDef, ch{ chDef = def - 1 }))
  , (fi spd, return (GSpd, ch{ chSpd = spd - 1 }))
  , (length swords, return (GSword (head swords), ch{ chSwords = tail swords }))
  ]
  where
    fi = fromInteger

weakenG :: Character -> Gen (Growth, Character)
weakenG ch@Ch{ chHp = hp, chAtk = atk, chDef = def, chSpd = spd, chSwords = swords } = join $ chooseRated
  [ (fi hp, return (GHP, ch{ chHp = hp - 5 }))
  , (fi atk, return (GAtk, ch{ chAtk = atk - 1 }))
  , (fi def, return (GDef, ch{ chDef = def - 1 }))
  , (fi spd, return (GSpd, ch{ chSpd = spd - 1 }))
  , (length swords, delete_sword)
  ]
  where
    delete_sword = do
      (s, ss) <- chooseSelect swords
      return (GSword s, ch{ chSwords = ss })
    fi = fromInteger

weakenings :: Character -> [(Character, Growth)]
weakenings ch@Ch{ chHp = hp, chAtk = atk, chDef = def, chSpd = spd, chSwords = swords } = join
  [ guard (hp > 0) >> return (ch{ chHp = hp - 5 }, GHP)
  , guard (atk > 0) >> return (ch{ chAtk = atk - 1 }, GAtk)
  , guard (def > 0) >> return (ch{ chDef = def - 1 }, GDef)
  , guard (spd > 0) >> return (ch{ chSpd = spd - 1 }, GSpd)
  , do
      (s, ss) <- selections swords
      return (ch{ chSwords = ss }, GSword s)
  ]

selections :: [a] -> [(a, [a])]
selections xs = do
  n <- [0..length xs-1]
  let (x, t:y) = splitAt n xs
  return (t, x ++ y)

m100Growth :: [Growth]
m100Growth = GHP : GAtk : GDef : GSpd : map GSword m100Swords

allGrowth :: [Sword] -> [Growth]
allGrowth ss = GHP : GAtk : GDef : GSpd : map GSword ss

charToChrom :: Character -> Gen Chrom
charToChrom = loop []
  where
    loop acc ch
      | all ((==0) . ($ ch)) [chHp, chAtk, chDef, chSpd, fromIntegral . length . chSwords] = return $ reverse acc
      | otherwise = do
        (k, v) <- weaken ch
        loop (k:acc) v

initialChar :: Character
initialChar = Ch
  { chName = ""
  , chHp = 0
  , chAtk = 0
  , chDef = 0
  , chSpd = 0
  , chSwords = []
  }

chooseM :: [Gen a] -> Gen a
chooseM bs = join $ choose bs

choose :: [a] -> Gen a
choose [] = error "choose []"
choose xs = do
  n <- rnd $ length xs
  return $ xs !! n

chooseSplit :: [a] -> Gen ([a], [a])
chooseSplit xs = do
  n <- rnd $ 1 + length xs
  return $ splitAt n xs

chooseSelectN :: Int -> [a] -> Gen ([a], [a])
chooseSelectN 0 xs = return ([], xs)
chooseSelectN n xs = do
  (x, r) <- chooseSelect xs
  (rr, rrr) <- chooseSelectN (n-1) r
  return (x:rr, rrr)

chooseReplace :: a -> [a] -> Gen [a]
chooseReplace x xs = do
  n <- rnd $ length xs
  let (a, _:b) = splitAt n xs
  return $ a ++ x:b

chooseSelect :: [a] -> Gen (a, [a])
chooseSelect xs = do
  n <- rnd $ length xs
  let (a, b:c) = splitAt n xs
  return (b, a++c)

chooseN :: Int -> [a] -> Gen [a]
chooseN n xs = f n (length xs) xs
  where
    f 0 _ _ = return []
    f n m [] = return []
    f n m (x:xs) = do
      v <- rnd m
      if v < n
        then liftM (x:) $ f (n-1) (m-1) xs
        else f n (m-1) xs

chooseRatedN :: Int -> [(Int, a)] -> Gen [a]
chooseRatedN k lst = loop (sum $ map fst lst) k lst
  where
    loop total 0 _ = return []
    loop total k [] = return []
    loop total k lst = do
      (x, rest, new_total) <- chooseRatedOne total lst
      m <- loop new_total (k-1) rest
      return (x:m)

chooseRated :: [(Int, a)] -> Gen a
chooseRated lst = do
  (x, _, _) <- chooseRatedOne (sum $ map fst lst) lst
  return x

chooseRatedOne :: Int -> [(Int, a)] -> Gen (a, [(Int, a)], Int)
chooseRatedOne total lst = do
  n <- rnd total
  case span ((<=n) . fst) $ zip threash lst of
    (x, (_, (r, p)):y) -> return (p, map snd $ x ++ y, total - r)
  where
    nums = map fst lst
    threash = scanl1 (+) nums

split3 :: Int -> Int -> [a] -> ([a], [a], [a])
split3 n m t = (a, b, c)
  where
    (a, bc) = splitAt n t
    (b, c) = splitAt (m-n) bc

cross :: Chrom -> Chrom -> Gen Chrom
cross x y = do
  n <- choose [0..len-1]
  m <- choose [n+1..len]
  let (x0, x1, x2) = split3 n m x
  let (y0, y1, y2) = split3 n m y
  return $ x0 ++ y1 ++ x2
  where
    len = length x

genCharByGrowth gg n = liftM chromToChar $ replicateM n gg

genChar15 = genCharByGrowth genGrowth15

genGrowth15 :: Gen Growth
genGrowth15 = genGrowthBySword genSword15

genSword15 :: Gen Sword
genSword15 = chooseM cs
  where
    cs = map return [ slash, rapid, heavy, shock, poison, absolute, magical, death, heat, protect, accel, shield, mirror, life] ++ [choose recoveries]
    recoveries = map (recovery$) [1..20]

m15swords :: [Sword]
m15swords = [ slash, rapid, heavy, shock, poison, absolute, magical, death, heat, protect, accel, shield, mirror, life] ++ map recovery [1..20]

shuffle :: [a] -> IO [a]
shuffle [] = return []
shuffle xs = do
  (x, z) <- chooseSelect xs
  zz <- shuffle z
  return $ x:zz

weakensG :: Character -> Int -> Gen (Character, [Growth])
weakensG ch 0 = return (ch, [])
weakensG ch n = do
  (ch1, k) <- choose $ weakenings ch
  (ch2, gg) <- weakensG ch1 (n-1)
  return (ch2, k:gg)

countGrowth :: Growth -> Character -> Integer
countGrowth g ch = case g of
  GHP -> div (chHp ch) 5
  GAtk -> chAtk ch
  GDef -> chDef ch
  GSpd -> chSpd ch
  GSword sw -> fromIntegral $ length $ filter (==sw) $ chSwords ch
