import Control.Concurrent(threadDelay)
import Control.Monad
import Control.Exception
import Data.List
import Data.Maybe
import System.IO.SSC.Haskell98 as SSC
import System.Random
import System.Process
import System.Environment
import System.Console.GetOpt
import System.Directory
import System.Exit
import System.FilePath
import System.IO.Error(isAlreadyExistsError, isDoesNotExistError, isPermissionError, ioeGetErrorType)
import GHC.IOBase(IOErrorType(Interrupted))

import Search
import Char
import Swords
import M100Util

data Settings
  = Se
  { seTaboo :: Maybe FilePath
  , seTimeout :: Double
  , seExact :: Bool
  , seSwordset :: SwordsetName
  }

data Mode = HelpMode | NormalMode Settings
data SwordsetName = BasicSS | M100SS | M2011aSS | M2011SS | MVoidSS

readSS :: String -> SwordsetName
readSS s = case s of
  "basic" -> BasicSS
  "m100" -> M100SS
  "m2011a" -> M2011aSS
  "m2011" -> M2011SS
  "mvoid" -> MVoidSS
  _ -> error $ "unknown swordset: " ++ show s

modifySe :: (Settings -> Settings) -> Mode -> Mode
modifySe f HelpMode = HelpMode
modifySe f (NormalMode x) = NormalMode (f x)

defaultSe :: Settings
defaultSe = Se 
  { seTaboo = Nothing
  , seTimeout = 1
  , seSwordset = M100SS
  , seExact = False
  }

options :: [OptDescr (Mode -> Mode)]
options =
  [ Option "a" ["avoid"] (ReqArg (\t -> modifySe $ \se -> se{ seTaboo = Just t }) "FILE")
      "specify avoid list"
  , Option "t" ["timeout"] (ReqArg (\t -> modifySe $ \se -> se{ seTimeout = read t }) "SECONDS")
      "specify simulator timeout, in fractional"
  , Option "e" ["exact"] (NoArg $ modifySe $ \se -> se{ seExact = True })
      "require contents of resut.txt to match exactly"
  , Option "s" ["swordset"] (ReqArg (\s -> modifySe $ \se -> se{ seSwordset = readSS s }) "SET")
      "specify sword set to use (available: basic, m100, m2011a, m2011, mvoid)"
  , Option "b" ["only-basic"] (NoArg $ modifySe $ \se -> se{ seSwordset = BasicSS })
      "use only the basic 15 swords (same as --swordset=basic)"
  , Option "h" ["help"] (NoArg $ const HelpMode) "show this help"
  ]

handleOptions :: [String] -> IO (FilePath, FilePath, Settings)
handleOptions args = case getOpt Permute options args of
  (ms, as, []) -> case foldr ($) (NormalMode defaultSe) ms of
    HelpMode -> show_help >> exitWith ExitSuccess
    NormalMode se -> case as of
      [e0, e1] -> return (e0, e1, se)
      _ -> do
        SSC.putStrLn "wrong number of arguments"
        show_usage >> exitWith (ExitFailure 1)
  (_, _, msgs) -> do
    mapM_ SSC.putStrLn msgs
    show_usage >> exitWith (ExitFailure 1)
  where
    show_usage = do
      SSC.putStrLn usage
      SSC.putStrLn "use -h for help"
    usage = "Usage: random-compare [OPTIONS...] CMD CMD"
    show_help = SSC.putStr $ usageInfo usage options

makeGoodChar gen taboo = do
  ch <- liftM chromToChar gen
  if any (flip elem taboo) $ show ch
    then makeGoodChar gen taboo
    else return ch

testSame :: IO Chrom -> Bool -> Double -> String -> FilePath -> FilePath -> IO ()
testSame gen exact timeout taboo x y = loop 0 Nothing
  where
    loop n outdir = do
      let iter = "iteration " ++ show n
      SSC.putStr $ '\r':iter
      SSC.hFlush SSC.stdout
      cs <- replicateM 5 $ makeGoodChar gen taboo
      let input = unlines $ map (process . show) $ uniquifyNames cs
      SSC.writeFile "player.txt" input
      r0 <- run timeout "0" x iter
      r1 <- run timeout "1" y iter
      comp <- if exact then compare_exact else return True
      outdir' <- if r0 == r1 && comp
        then do
          mapM_ (ignore isDoesNotExistError . removeFile) files
          return outdir
        else do
          SSC.putStr "\r"
          outdir' <- case outdir of
            Just k -> return k
            Nothing -> newDirectory "random-record-"
          let dir = outdir' </> show n
          createDirectory dir
          let move_to_dir f = ignore isDoesNotExistError $ renameFile f (dir </> f)
          mapM_ move_to_dir files
          SSC.putStrLn $ "mismatch found: " ++ show cs
          SSC.putStrLn $ "details saved to " ++ dir
          return $ Just outdir'
      loop (n + 1) outdir'
    process = mapMaybe filt . dropWhile (==' ')
    filt 'ｺ' = Just 'コ'
    filt 'ｳ' = Nothing
    filt '.' = Nothing
    filt c = Just c
    files = ["0-roundrobin.txt", "0-result.txt", "1-roundrobin.txt", "1-result.txt", "player.txt"]

compare_exact :: IO Bool
compare_exact = do
  s0 <- get "0-result.txt"
  s1 <- get "1-result.txt"
  return $ s0 == s1
  where
    get f = liftM (either (const "") id) $ tryJust (guard . isDoesNotExistError) $ SSC.readFile f

main = do
  (c0, c1, se) <- handleOptions =<< getArgs
  taboo <- case seTaboo se of
    Just k -> SSC.readFile k
    _ -> return ""
  testSame (generator $ seSwordset se) (seExact se) (seTimeout se) taboo c0 c1

generator :: SwordsetName -> Gen Chrom
generator BasicSS = replicateM 10 genGrowth15
generator M100SS = genChrom
generator M2011aSS = genChromG m2011aSS
generator M2011SS = genChromG m2011SS
generator MVoidSS = genChromG mvoidSS

newDirectory :: FilePath -> IO FilePath
newDirectory prefix = loop 0
  where
    loop n = do
      r <- tryJust (guard . isAlreadyExistsError) $ createDirectory path
      case r of
        Left _ -> loop $ n + 1
        Right _ -> return path
      where
        path = prefix ++ show n

run :: Double -> String -> FilePath -> String -> IO [[Int]]
run timeout pfx cmd desc = do
  good <- runCmd timeout cmd
  result <- if good
    then do
      renameFile "roundrobin.txt" rr
      ignore isDoesNotExistError $ renameFile "result.txt" $ pfx ++ "-result.txt"
      readResult rr
    else do
      SSC.writeFile rr "timed out\n"
      return []
  return result
  where
    rr = pfx ++ "-roundrobin.txt"

ignore :: (IOException -> Bool) -> IO a -> IO ()
ignore cond x = do
  tryJust (guard . cond) x
  return ()

runCmd :: Double -> FilePath -> IO Bool
runCmd timeout cmd = do
  (_, _, _, ph) <- createProcess $ proc cmd []
  exited <- wait count ph
  when (not exited) $ do
    ignore isPermissionError $ terminateProcess ph
    exited1 <- wait 100 ph
    when (not exited1) $ fail "couldn't terminate a process"
  return exited
  where
    wait 0 ph = return False
    wait n ph = do
      threadDelay (10*1000)
      code <- tryJust (guard . (==Interrupted) . ioeGetErrorType) $ getProcessExitCode ph
      case code of
        Right Nothing -> wait (n-1) ph
        _ -> return True
    count = round $ timeout * 100

readResult :: FilePath -> IO [[Int]]
readResult result = do
  r <- tryJust (guard . isDoesNotExistError) $ SSC.openFile result ReadMode
  case r of
    Left _ -> return []
    Right h -> do_read h `finally` SSC.hClose h
  where
    do_read h = do
      loop h
      ls <- loop1 h
      return $ parseResult ls
    loop h = do
      eof <- hIsEOF h
      when (not eof) $ do
        ln <- hGetLine h
        when (ln /= "RESULT") $ loop h
    loop1 h = do
      eof <- hIsEOF h
      if eof
        then return []
        else do
        ln <- hGetLine h
        rest <- loop1 h
        return (ln:rest)

parseResult :: [String] -> [[Int]]
parseResult = map (mapMaybe parseR) . filter (elem "++") . map words
  where
    parseR "@@" = Just 2
    parseR "++" = Just 1
    parseR "." = Just 0
    parseR _ = Nothing
