{-# OPTIONS_GHC -fglasgow-exts #-}
{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental

  Abstract representation of opened file. This module handles what
  are called regular files on Unix.
-}
module System.IO.File
  ( File
  , RFile
  , WFile
  , RWFile
  , fOpen
  , fReadOpen
  , fWriteOpen
  , fAppendOpen
  , fReadWriteOpen
  , fOpenTemp
  , fClose
  , fRead
  , fReadNonBlocking
  , fWrite
  , fWriteNonBlocking
  , fSeek
  , fTell
  , fReady
  , fFileSize
  , fSetFileSize
  , fSync
  , fWithFd
  ) where

import Control.Monad
import Control.Concurrent.MVar
import Foreign.Ptr
import Data.Yesno
import System.IO.Fd
import System.IO(IOMode(..), SeekMode(..))

-- | File object. The `r` parameter indicates whether the file is
-- opened for reading. If the `r` parameter is `Data.Yesno.YesTag`,
-- that means the file is opened for reading, and if it `r` is
-- `Data.Yesno.NoTag`, the file is not for reading. Similarly, the
-- `w` parameter indicates whether the file is opened for writing.
newtype File r w = File (MVar Fd)
type RFile = File YesTag NoTag
type WFile = File NoTag YesTag
type RWFile = File YesTag YesTag

fOpen :: FilePath -> IOMode -> IO (File r w)
fOpen path mode = do
  fd <- openRawBinaryFd path mode
  v <- newMVar fd
  return $ File v

fReadOpen :: FilePath -> IO RFile
fReadOpen path = fOpen path ReadMode

fWriteOpen :: FilePath -> IO WFile
fWriteOpen path = fOpen path WriteMode

fAppendOpen :: FilePath -> IO WFile
fAppendOpen path = fOpen path AppendMode

fReadWriteOpen :: FilePath -> IO RWFile
fReadWriteOpen path = fOpen path ReadWriteMode

fOpenTemp :: FilePath -> String -> IO (FilePath, RWFile)
fOpenTemp tmp_dir template = do
  (path, fd) <- openTempFd tmp_dir template True
  v <- newMVar fd
  return $ (path, File v)

fClose :: File r w -> IO ()
fClose (File rfd) = withMVar rfd fdClose

fRead :: File YesTag w -> Ptr a -> Int -> IO Int
fRead (File rfd) buf size = withMVar rfd $ \fd -> fdGetBuf fd buf size

fReadNonBlocking :: File YesTag w -> Ptr a -> Int -> IO (Maybe Int)
fReadNonBlocking (File rfd) buf size = withMVar rfd $ \fd -> fdGetBufNonBlocking fd buf size

fWrite :: File r YesTag -> Ptr a -> Int -> IO Int
fWrite (File rfd) buf size = withMVar rfd $ \fd -> fdPutBuf fd buf size

fWriteNonBlocking :: File r YesTag -> Ptr a  -> Int -> IO (Maybe Int)
fWriteNonBlocking (File rfd) buf size = withMVar rfd $ \fd -> fdPutBufNonBlocking fd buf size

fSeek :: File r w -> SeekMode -> Integer -> IO ()
fSeek (File rfd) mode pos = withMVar rfd $ \fd -> fdSeek fd mode pos

fTell :: File r w -> IO Integer
fTell (File rfd) = withMVar rfd fdTell

fReady :: File YesTag w -> IO Bool
fReady (File rfd) = withMVar rfd $ \fd -> fdReady fd False

fFileSize :: File r w -> IO Integer
fFileSize (File rfd) = withMVar rfd $ \fd -> fdFileSize fd

fSetFileSize :: File r YesTag -> Integer -> IO ()
fSetFileSize (File rfd) size = withMVar rfd $ \fd -> fdSetFileSize fd size

fSync :: File r YesTag -> IO ()
fSync (File rfd) = withMVar rfd fdSync

fWithFd :: File r w -> (Fd -> IO a) -> IO (Maybe a)
fWithFd (File rfd) f = liftM Just $ withMVar rfd f
