{-# OPTIONS_GHC -fglasgow-exts #-}
{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental

  Generic transformer for buffer-based conversion. This transformer
  provides block IO port operations on top of memory IO port, using
  user-supplied conversion routine.
-}
module System.IO.SSC.Convert
  ( Converter
  , identityConverter
  , iConversionBlock
  , oConversionBlock
  , ioConversionBlock
  , IConversionBlock
  , OConversionBlock
  , IOConversionBlock
  , ConversionPos
  , converterWithFallback
  ) where

import Control.Monad
import Data.Word
import Foreign.Marshal.Utils
import Foreign.Storable
import Foreign.Marshal.Array
import Foreign.Ptr

import System.IO.SSC.Classes
import System.IO.SSC.Internal.Ptr
import System.IO.SSC.Internal.Debug

-- | Type of the conversion routine.
--
-- A conversion function `f` is
-- used in the form of `f` `ip` `is` `op` `os`, where `ip` is
-- pointer to the input buffer, `is` is size (in elements) of the
-- input buffer, `op` is pointer to the output buffer, and `os` is
-- size of the output buffer. When called, the routine should perform
-- the conversion and return a pair consisting of the number of
-- elements read and the number of elements written.
--
-- For efficiency,
-- the routine should continue conversion until either the input
-- buffer is exhausted or the output buffer is full.
type Converter from to = 
  Ptr from     -- ^ pointer to the input buffer
  -> Int       -- ^ size (in elements) of the input buffer
  -> Ptr to    -- ^ pointer to the output buffer
  -> Int       -- ^ size (in elements) of the output buffer
  -> IO (Int, Int) -- ^ (number of elements read, number of elements written)

unPtr :: Ptr a -> a
unPtr = undefined

-- | A converter that just copies the input to the output.
identityConverter :: (Storable a) => Converter a a
identityConverter ip is op os = do
  copyBytes op ip sz
  return (n, n)
  where
    n = min is os
    sz = sizeOf $ unPtr ip

data IOConversionBlock d pos m a base top = IOCB m (Converter base top) (Converter top base)
data IConversionBlock d pos m a base top = ICB m (Converter base top)
data OConversionBlock d pos m a base top = OCB m (Converter top base)

newtype ConversionPos pos = CP pos

iConversionBlock :: Converter base top -> m -> IO (IConversionBlock d pos m a base top)
iConversionBlock up base = return $ ICB base up

oConversionBlock :: Converter top base -> m -> IO (OConversionBlock d pos m a base top)
oConversionBlock dn base = return $ OCB base dn

ioConversionBlock :: Converter base top -> Converter top base -> m -> IO (IOConversionBlock d pos m a base top)
ioConversionBlock up down base = return $ IOCB base up down

instance (Port m) => Port (IOConversionBlock d pos m a base top) where
  rawClose (IOCB m _ _) = rawClose m
  rawIsClosed (IOCB m _ _) = rawIsClosed m
  rawFlush (IOCB m _ _) = rawFlush m

instance (Port m) => Port (IConversionBlock d pos m a base top) where
  rawClose (ICB m _) = rawClose m
  rawIsClosed (ICB m _) = rawIsClosed m
  rawFlush (ICB m _) = rawFlush m

instance (Port m) => Port (OConversionBlock d pos m a base top) where
  rawClose (OCB m _) = rawClose m
  rawIsClosed (OCB m _) = rawIsClosed m
  rawFlush (OCB m _) = rawFlush m

instance (Destination d m) => Destination d (IOConversionBlock d pos m a base top) where
  rawDest (IOCB m _ _) = rawDest m

instance (Destination d m) => Destination d (IConversionBlock d pos m a base top) where
  rawDest (ICB m _) = rawDest m

instance (Destination d m) => Destination d (OConversionBlock d pos m a base top) where
  rawDest (OCB m _) = rawDest m

instance (Lockable m) => Lockable (IOConversionBlock d pos m a base top) where
  sLock (IOCB m _ _) = sLock m
  sTryLock (IOCB m _ _) = sTryLock m
  sUnsafeUnlock (IOCB m _ _) = sUnsafeUnlock m

instance (Lockable m) => Lockable (IConversionBlock d pos m a base top) where
  sLock (ICB m _) = sLock m
  sTryLock (ICB m _) = sTryLock m
  sUnsafeUnlock (ICB m _) = sUnsafeUnlock m

instance (Lockable m) => Lockable (OConversionBlock d pos m a base top) where
  sLock (OCB m _) = sLock m
  sTryLock (OCB m _) = sTryLock m
  sUnsafeUnlock (OCB m _) = sUnsafeUnlock m

instance (Discardable m) => Discardable (IOConversionBlock d pos m a base top) where
  rawDiscard (IOCB m _ _) = rawDiscard m

instance (Discardable m) => Discardable (IConversionBlock d pos m a base top) where
  rawDiscard (ICB m _) = rawDiscard m

instance (Discardable m) => Discardable (OConversionBlock d pos m a base top) where
  rawDiscard (OCB m _) = rawDiscard m

instance (Seekable pos m) => Seekable (ConversionPos pos) (IConversionBlock d pos m a base top) where
  rawGetPos (ICB m _) = liftM CP $ rawGetPos m
  rawSetPos (ICB m _) (CP pos) = rawSetPos m pos

instance (Seekable pos m) => Seekable (ConversionPos pos) (OConversionBlock d pos m a base top) where
  rawGetPos (OCB m _) = liftM CP $ rawGetPos m
  rawSetPos (OCB m _) (CP pos) = rawSetPos m pos

instance (Seekable pos m) => Seekable (ConversionPos pos) (IOConversionBlock d pos m a base top) where
  rawGetPos (IOCB m _ _) = liftM CP $ rawGetPos m
  rawSetPos (IOCB m _ _) (CP pos) = rawSetPos m pos

getBuf :: (IMemory base m, Storable top) => BlockingMode -> m -> Converter base top -> Ptr top -> Int -> IO (Int, Bool)
getBuf bmode mem conv optr osize = do
  msg $ "+getBuf osize=" ++ show osize
  (begin, end) <- rawReadMem mem
  r <- if begin == end
    then do
      rawEndRead mem begin
      r <- rawFillMemB mem initBmode
      case r of
        Nothing -> return (0, True)
        Just 0 -> return (0, False)
        Just _ -> do
          (begin', end') <- rawReadMem mem
          n <- get 0 begin' end'
          loop n
    else do
      n <- get 0 begin end
      loop n
  msg $ "-getBuf r=" ++ show r
  return r
  where
    get n begin end = do
      (read, written) <- conv begin (distancePtr end begin) (advancePtr optr n) (osize - n)
      msg $ "convert read=" ++ show read ++ "; written=" ++ show written
      rawEndRead mem (advancePtr begin read)
      return written

    loop n 
      | n == osize = return (n, False)
      | otherwise = do
      r <- rawFillMemB mem contBmode
      case r of
        Nothing -> return (n, True)
        Just 0 -> return (n, False)
        Just _ -> do
          (begin, end) <- rawReadMem mem
          p <- get n begin end
          loop (n + p)

    (initBmode, contBmode) = case bmode of
      Blocking -> (NonBlockingPreferred, NonBlockingPreferred)
      NonBlockingPreferred -> (NonBlockingPreferred, NonBlocking)
      NonBlocking -> (NonBlocking, NonBlocking)

instance (IMemory base m, Storable top) => IBlock top (IOConversionBlock d pos m a base top) where
  rawGetBufB (IOCB m up down) bmode = getBuf bmode m up

instance (IMemory base m, Storable top) => IBlock top (IConversionBlock d pos m a base top) where
  rawGetBufB (ICB m up) bmode = getBuf bmode m up

putBuf :: (OMemory base m, Storable top) => BlockingMode -> m -> Converter top base -> Ptr top -> Int -> IO (Int, Bool)
putBuf bmode mem conv iptr isize = do
  msg $ "+putBuf isize=" ++ show isize
  (begin, end) <- rawWriteMem mem
  r <- if begin == end
    then do
      rawEndWrite mem begin
      r <- rawFlushMemB mem initBmode
      case r of
        (_, True) -> return (0, True)
        (0, False) -> return (0, False)
        (_, False) -> do
          (begin', end') <- rawWriteMem mem
          n <- put 0 begin' end'
          loop n
    else do
      n <- put 0 begin end
      loop n
  msg $ "-putBuf; r=" ++ show r
  return r
  where
    put n begin end = do
      (read, written) <- conv (advancePtr iptr n) (isize - n) begin (distancePtr end begin)
      rawEndWrite mem (advancePtr begin written)
      return read

    loop n
      | n == isize = return (n, False)
      | otherwise = do
      r <- rawFlushMemB mem contBmode
      case r of
        (_, True) -> return (n, True)
        (0, False) -> return (n, False)
        (_, False) -> do
          (begin, end) <- rawWriteMem mem
          p <- put n begin end
          loop (n + p)
    (initBmode, contBmode) = case bmode of
      Blocking -> (NonBlockingPreferred, NonBlockingPreferred)
      NonBlockingPreferred -> (NonBlockingPreferred, NonBlocking)
      NonBlocking -> (NonBlocking, NonBlocking)

instance (OMemory base m, Storable top) => OBlock top (IOConversionBlock d pos m a base top) where
  rawPutBufB (IOCB mem up down) bmode = putBuf bmode mem down

instance (OMemory base m, Storable top) => OBlock top (OConversionBlock d pos m a base top) where
  rawPutBufB (OCB mem down) bmode = putBuf bmode mem down

instance (IMemory base m) => IMemory base (IConversionBlock d pos m a base top) where
  rawReadMem (ICB mem _) = rawReadMem mem
  rawEndRead (ICB mem _) = rawEndRead mem
  rawFillMemB (ICB mem _) = rawFillMemB mem

instance (IMemory base m) => IMemory base (IOConversionBlock d pos m a base top) where
  rawReadMem (IOCB mem _ _) = rawReadMem mem
  rawEndRead (IOCB mem _ _) = rawEndRead mem
  rawFillMemB (IOCB mem _ _) = rawFillMemB mem

instance (OMemory base m) => OMemory base (OConversionBlock d pos m a base top) where
  rawWriteMem (OCB mem _) = rawWriteMem mem
  rawEndWrite (OCB mem _) = rawEndWrite mem
  rawFlushMemB (OCB mem _) = rawFlushMemB mem

instance (OMemory base m) => OMemory base (IOConversionBlock d pos m a base top) where
  rawWriteMem (IOCB mem _ _) = rawWriteMem mem
  rawEndWrite (IOCB mem _ _) = rawEndWrite mem
  rawFlushMemB (IOCB mem _ _) = rawFlushMemB mem

converterWithFallback
  :: (Storable a, Storable b)
  => (Ptr a -> Int -> Ptr b -> Int -> IO (Int, Int, Bool))
  -> (Ptr a -> Int -> IO (Int, [a]))
  -> Converter a b
converterWithFallback main fallback iptr isize optr osize = loop 0 0
  where
    loop m n
      | m == isize || n == osize = return (m, n)
      | otherwise = do
        (read, written, succeed) <- main (advancePtr iptr m) (isize - m) (advancePtr optr n) (osize - n)
        let m' = m + read
        let n' = n + written
        if succeed
          then return (m', n')
          else do
            (nc, alt) <- fallback (advancePtr iptr m') (isize - m')
            if nc == 0
              then return (m', n')
              else do
                (m'', n'', cont) <- withArrayLen alt $ \len palt -> do
                  (read', written', succeed') <- main palt len (advancePtr optr n') (osize - n')
                  if not succeed'
                    then ioError $ userError "System.IO.SSC.Convert: Fallback failed"
                    else if read' < len
                      then return (m', n', False)
                      else return (m' + nc, n' + written', True)
                if cont
                  then loop m'' n''
                  else return (m'', n'')

instance (IStream a m) => IStream a (IConversionBlock d pos m a base top) where
  rawGet (ICB mem _) = rawGet mem
  rawGets (ICB mem _) = rawGets mem
  rawGetContents (ICB mem _) = rawGetContents mem

instance (IStream a m) => IStream a (IOConversionBlock d pos m a base top) where
  rawGet (IOCB mem _ _) = rawGet mem
  rawGets (IOCB mem _ _) = rawGets mem
  rawGetContents (IOCB mem _ _) = rawGetContents mem

instance (RetrievableStream a m) => RetrievableStream a (IConversionBlock d pos m a base top) where
  rawGetUntil (ICB mem _) = rawGetUntil mem

instance (RetrievableStream a m) => RetrievableStream a (IOConversionBlock d pos m a base top) where
  rawGetUntil (IOCB mem _ _) = rawGetUntil mem

instance (OStream a m) => OStream a (OConversionBlock d pos m a base top) where
  rawPut (OCB mem _) = rawPut mem
  rawPuts (OCB mem _) = rawPuts mem

instance (OStream a m) => OStream a (IOConversionBlock d pos m a base top) where
  rawPut (IOCB mem _ _) = rawPut mem
  rawPuts (IOCB mem _ _) = rawPuts mem

