{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental

   An implementation of Haskell98 IO library on top of SSC.

   This module also provides many of GHC extensions, but there are
   some differences:
  
   * Files opened in text mode is assumed to be encoded in the 
   locale-dependent encoding scheme, not latin-1.

   * In addition to binary mode and text mode, there is latin-1 mode
   which provides faster access to latin-1 files.

   * Text handles cannot be in read-write mode. Text handles and
   latin-1 handles are not seekable.

   * stdout is not automatically flushed at program exit.

   * hSetBuffering, hSetBinaryMode, and hSetLatin1Mode can only
   operate on handles just created. Once it is used, buffering
   mode and conversion mode can no longer be changed.

   * hGetBuf and other binary operations cannot be applied to text
   mode handles.

   * openTempFile is a synomym for openLatin1TempFile.
-}
module System.IO.SSC.Haskell98
  ( IO
  , fixIO
  , FilePath
  , Handle
  , stdin
  , stdout
  , stderr
  , openFile
  , IOMode(..)
  , hClose
  , readFile
  , writeFile
  , appendFile
  , hFileSize
  , hSetFileSize
  , hIsEOF
  , isEOF
  , BufferMode(..)
  , hSetBuffering
  , hGetBuffering
  , hFlush
  , hGetPosn
  , hSetPosn
  , HandlePosn
  , hSeek
  , SeekMode(..)
  , hTell
  , hIsOpen
  , hIsClosed
  , hIsReadable
  , hIsWritable
  , hIsSeekable
  , hIsTerminalDevice
  , hSetEcho
  , hGetEcho
  , hShow
  , hWaitForInput
  , hReady
  , hGetChar
  , hGetLine
  , hLookAhead
  , hGetContents
  , hPutChar
  , hPutStr
  , hPutStrLn
  , hPrint
  , interact
  , putChar
  , putStr
  , putStrLn
  , print
  , getChar
  , getLine
  , getContents
  , readIO
  , readLn
  , openBinaryFile
  , hSetBinaryMode
  , openLatin1File
  , hSetLatin1Mode
  , hPutBuf
  , hGetBuf
  , hPutBufNonBlocking
  , hGetBufNonBlocking
  , openTempFile
  , openBinaryTempFile
  , openLatin1TempFile
  ) where

import System.IO(fixIO, BufferMode(..), IOMode(..), SeekMode(..))
import System.IO.SSC.Haskell98.Base
import Prelude hiding (putChar, putStr, putStrLn, print, getChar,
  getLine, getContents, interact, readFile, writeFile, appendFile, readLn)

readFile :: FilePath -> IO String
readFile fname = do
  h <- openFile fname ReadMode
  hGetContents h

writeFile :: FilePath -> String -> IO ()
writeFile fname content = do
  h <- openFile fname WriteMode
  hPutStr h content
  hClose h

appendFile :: FilePath -> String -> IO ()
appendFile fname content = do
  h <- openFile fname AppendMode
  hPutStr h content
  hClose h

isEOF :: IO Bool
isEOF = hIsEOF stdin

hGetPosn :: Handle -> IO HandlePosn
hGetPosn h = do
  pos <- hTell h
  return $ HP $ hSeek h AbsoluteSeek pos 

hSetPosn :: HandlePosn -> IO ()
hSetPosn (HP x) = x

newtype HandlePosn = HP (IO ())

hReady :: Handle -> IO Bool
hReady h = hWaitForInput h 0

hPrint :: (Show a) => Handle -> a -> IO ()
hPrint h x = hPutStrLn h (show x)

interact :: (String -> String) -> IO ()
interact f = do
  c <- getContents
  putStr (f c)

putChar :: Char -> IO ()
putChar = hPutChar stdout

putStr :: String -> IO ()
putStr = hPutStr stdout

putStrLn :: String -> IO ()
putStrLn = hPutStrLn stdout

print :: (Show a) => a -> IO ()
print = hPrint stdout

getChar :: IO Char
getChar = hGetChar stdin

getLine :: IO String
getLine = hGetLine stdin

getContents :: IO String
getContents = hGetContents stdin

readLn :: (Read a) => IO a
readLn = getLine >>= readIO
