{-# OPTIONS_GHC -fglasgow-exts #-}
{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental

   Internal structure of Haskell98 emulation.
-}
module System.IO.SSC.Haskell98.Base where

import Control.Monad
import Control.Concurrent.MVar
import Data.Word
import Data.Maybe
import Data.Yesno
import Data.Typeable
import Foreign.Ptr
import System.Posix.Types(Fd)
import System.IO(BufferMode(..), IOMode(..), SeekMode(..))
import System.IO.Error
import System.IO.Unsafe
import System.IO.File

import System.IO.SSC.Internal.Fd
import System.IO.SSC.Internal.Char

import System.IO.SSC.Classes
import System.IO.SSC.Buffer
import System.IO.SSC.Latin1Scanner
import System.IO.SSC.TextMode
import System.IO.SSC.LocaleEncoding
import System.IO.SSC.Scanner
import System.IO.SSC.BufferSize
import System.IO.SSC.AutoFlush
import System.IO.SSC.LineBuffering
import System.IO.SSC.FilePort
import System.IO.SSC.Utils
import System.IO.SSC.Stdio(rawStdin, rawStdout, rawStderr)

data Handle = Handle 
  { haBody :: !(MVar HBody)
  , haName :: String
  }
  deriving (Typeable)

instance Show Handle where
  show a = "{handle:" ++ haName a ++ "}"

instance Eq Handle where
  a == b = haBody a == haBody b

data ConversionMode
  = BinaryMode
  | TextMode
  | Latin1Mode
  deriving (Show, Eq)

data HBody
  = HOpen 
    { haReader :: Maybe TextReader
    , haBinReader :: Maybe BinaryReader
    , haWriter :: Maybe TextWriter
    , haBinWriter :: Maybe BinaryWriter
    , haSeeker :: Maybe Seeker
    , haFd :: Maybe Fd
    , haBufMode :: BufferMode
    , haConvMode :: ConversionMode
    }
  | HUninitialized 
    { haUninit :: Uninitialized
    , haBufMode :: BufferMode
    , haConvMode :: ConversionMode
    , haFd :: Maybe Fd
    }
  | HSemiClosed
    { haSemiClosedReader :: TextReader
    , haFd :: Maybe Fd
    }
  | HClosed

data Uninitialized
  = forall a s. (IBlock Word8 s) => UR s
  | forall s. (OBlock Word8 s) => UW s
  -- no URW constructer here beause rw ports needs to be seekable to be buffered
  | forall a s. (IBlock Word8 s, RelativeSeekable Integer s, SizedInput Integer s) => URS s
  | forall s. (OBlock Word8 s, RelativeSeekable Integer s, SizedInput Integer s) => UWS s
  | forall a s. (IBlock Word8 s, OBlock Word8 s, RelativeSeekable Integer s, SizedInput Integer s) => URWS s

data TextReader = forall a s. (RetrievableStream Char s, IMemory a s) 
  => TR s (IO Char) -- We need hLookAhead as an extra operation
data TextWriter = forall s. (OStream Char s) => TW s
data BinaryReader = forall s. (IBlock Word8 s) => BR s
data BinaryWriter = forall s. (OBlock Word8 s) => BW s
data Seeker = forall s. (RelativeSeekable Integer s, SizedInput Integer s) => Se s

type Abilities = (Maybe TextReader, Maybe TextWriter, Maybe BinaryReader, Maybe BinaryWriter, Maybe Seeker)

closeUninitialized :: Uninitialized -> IO ()
closeUninitialized u = case u of
  UR s -> rawClose s
  UW s -> rawClose s
  URS s -> rawClose s
  UWS s -> rawClose s
  URWS s -> rawClose s

justTrC s = Just (TR s (sLookAhead s))
justTrB s = Just (TR s (liftM b2c (sLookAhead s)))

fromR s = (justTrC s, Nothing, Nothing, Nothing, Nothing)
fromW s = (Nothing, Just (TW s), Nothing, Nothing, Nothing)
fromRW s = (justTrC s, Just (TW s), Nothing, Nothing, Nothing)
fromRB s = (justTrB s, Nothing, Just (BR s), Nothing, Nothing)
fromWB s = (Nothing, Just (TW s), Nothing, Just (BW s), Nothing)
fromRWB s = (justTrB s, Just (TW s), Just (BR s), Just (BW s), Nothing)
fromRS s = (justTrC s, Nothing, Nothing, Nothing, Just (Se s))
fromWS s = (Nothing, Just (TW s), Nothing, Nothing, Just (Se s))
fromRWS s = (justTrC s, Just (TW s), Nothing, Nothing, Just (Se s))
fromRBS s = (justTrB s, Nothing, Just (BR s), Nothing, Just (Se s))
fromWBS s = (Nothing, Just (TW s), Nothing, Just (BW s), Just (Se s))
fromRWBS s = (justTrB s, Just (TW s), Just (BR s), Just (BW s), Just (Se s))

initialize :: Uninitialized -> BufferMode -> ConversionMode -> Maybe Fd -> IO HBody
initialize u bmode cmode mfd = do
  when (cmode == TextMode) ensureLocaleInitialized
  (tr, tw, br, bw, s) <- case u of
    UR s -> mkAbilitiesR fromRB fromR fromRB s
    UW s -> mkAbilitiesW fromWB fromWB fromWB fromW fromW fromW fromWB fromWB fromWB s
    URS s -> mkAbilitiesR fromRBS fromR fromRB s
    UWS s -> mkAbilitiesW fromWBS fromWBS fromWBS fromW fromW fromW fromWB fromWB fromWB s
    URWS s -> case cmode of
      BinaryMode -> do
        s' <- return s
          >>= ioSizedBuffer size
          >>= ioLatin1Scanner
        case bmode of
          NoBuffering -> liftM fromRWBS $ ioAutoFlush s'
          LineBuffering -> liftM fromRWBS $ ioLineBuffering s'
          BlockBuffering _ -> return $ fromRWBS s'
      TextMode -> invariantViolation "initialize" "URWS/TextMode"
      Latin1Mode -> do
        s' <- return s
          >>= ioSizedBuffer esize
          >>= iorTextModeWithSize size
          >>= ioLatin1Scanner
        case bmode of
          NoBuffering -> liftM fromRWB $ ioAutoFlush s'
          LineBuffering -> liftM fromRWB $ ioLineBuffering s'
          BlockBuffering _ -> return $ fromRWB s'
  return HOpen
    { haReader = tr
    , haWriter = tw
    , haBinReader = br
    , haBinWriter = bw
    , haSeeker = s
    , haFd = mfd
    , haBufMode = bmode
    , haConvMode = cmode
    }
  where
    mkAbilitiesR mkB mkT mkL s = case cmode of
      BinaryMode -> liftM mkB $ 
        return s
          >>= iSizedBuffer size
          >>= iLatin1Scanner
      TextMode -> liftM mkT $
        return s
          >>= iSizedBuffer esize
          >>= iTextModeWithSize esize
          >>= maybe iLocaleEncoding iNaturalEncoding mfd
          >>= iSizedBuffer size
          >>= iScanner
      Latin1Mode -> liftM mkL $
        return s
          >>= iSizedBuffer esize
          >>= iTextModeWithSize size
          >>= iLatin1Scanner

    mkAbilitiesW mkBN mkBL mkBB mkTN mkTL mkTB mkLN mkLL mkLB s = case cmode of
      BinaryMode -> do
        s' <- return s
          >>= oSizedBuffer size
          >>= oLatin1Scanner
        case bmode of
          NoBuffering -> liftM mkBN $ oAutoFlush s'
          LineBuffering -> liftM mkBL $ oLineBuffering s'
          BlockBuffering _ -> return $ mkBB s'
      TextMode -> do
        s' <- return s
          >>= oSizedBuffer esize
          >>= oTextModeWithSize esize
          >>= maybe oLocaleEncoding oNaturalEncoding mfd
          >>= oSizedBuffer size
          >>= oScanner
        case bmode of
          NoBuffering -> liftM mkTN $ oAutoFlush s'
          LineBuffering -> liftM mkTL $ oLineBuffering s'
          BlockBuffering _ -> return $ mkTB s'
      Latin1Mode -> do
        s' <- return s
          >>= oSizedBuffer esize
          >>= oTextModeWithSize size
          >>= oLatin1Scanner
        case bmode of
          NoBuffering -> liftM mkLN $ oAutoFlush s'
          LineBuffering -> liftM mkLL $ oLineBuffering s'
          BlockBuffering _ -> return $ mkLB s'

    size = case bmode of
      NoBuffering -> defaultBufferSize
      LineBuffering -> defaultBufferSize
      BlockBuffering Nothing -> defaultBufferSize
      BlockBuffering (Just n) -> max 1 n
    esize = max minConvBufSize size

minConvBufSize :: Int
minConvBufSize = 100

invariantViolation :: String -> String -> IO a
invariantViolation pos desc = ioError $ userError msg
  where
    msg = "Invariant violation: System.IO.SSC.Haskell98.Base." ++ pos ++ ": " ++ desc

illegalOp :: String -> String -> Handle -> IO a
illegalOp pos desc h = illegalOpWithFname pos desc (Just (haName h))

illegalOpWithFname :: String -> String -> Maybe String -> IO a
illegalOpWithFname pos desc fname = ioError $ mkIOError illegalOperationErrorType ("System.IO.SSC.Haskell98.Base." ++ pos) Nothing fname

errorClosed :: String -> Handle -> IO a
errorClosed pos h = illegalOp pos "handle is closed" h

errorSemiClosed :: String -> Handle -> IO a
errorSemiClosed pos h = illegalOp pos "handle is semi-closed" h

errorUsed :: String -> Handle -> IO a
errorUsed pos h = illegalOp pos "handle already has been used" h

initLoc :: MVar Bool
initLoc = unsafePerformIO $ newMVar False

ensureLocaleInitialized :: IO ()
ensureLocaleInitialized = do
  initialized <- takeMVar initLoc
  when (not initialized) $ initCtypeLocale `catch` \e -> putMVar initLoc False >> ioError e
  putMVar initLoc True

mkFromU :: Uninitialized -> Maybe Fd -> ConversionMode -> HBody
mkFromU u mfd cmode = HUninitialized
  { haUninit = u
  , haFd = mfd
  , haConvMode = cmode
  , haBufMode = BlockBuffering (Just defaultBufferSize)
  }

mkFromFile :: File r w -> (FilePort r w -> Uninitialized) -> ConversionMode -> IO HBody
mkFromFile f mk cmode = do
  p <- filePort f
  mfd <- fWithFd f return
  return $ mkFromU (mk p) mfd cmode

withHandleAlt :: Handle -> (HBody -> IO (a, HBody)) -> IO a
withHandleAlt Handle{ haBody = bref } updater = do
  body <- takeMVar bref
  (r, newbody) <- updater body
    `catch` \e -> putMVar bref body >> ioError e
  putMVar bref newbody
  return r

withHandle :: Handle -> (HBody -> IO a) -> IO a
withHandle h a = withHandleAlt h (\b -> do r <- a b; return (r, b))

withHOpen 
  :: String
  -> Handle
  -> (Maybe TextReader
   -> Maybe TextWriter
   -> Maybe BinaryReader
   -> Maybe BinaryWriter
   -> Maybe Seeker
   -> Maybe Fd
   -> BufferMode
   -> IO a)
  -> IO a
withHOpen locname h a = withHandleAlt h $ \b -> case b of
  HOpen { haReader = r, haWriter = w, haBinReader = br, haBinWriter = bw, haSeeker = s, haFd = mfd, haBufMode = bmode } -> do
    r <- a r w br bw s mfd bmode
    return (r, b)
  HUninitialized { haUninit = u, haFd = mfd, haBufMode = bmode, haConvMode = cmode } -> do
    nb <- initialize u bmode cmode mfd
    r <- a
      (haReader nb)
      (haWriter nb)
      (haBinReader nb)
      (haBinWriter nb)
      (haSeeker nb)
      (haFd nb)
      (haBufMode nb)
    return (r, nb)
  HSemiClosed {} -> errorSemiClosed locname h
  HClosed {} -> errorClosed locname h

newHandle :: String -> HBody -> IO Handle
newHandle name body = do
  bref <- newMVar body
  return $ Handle
    { haName = name
    , haBody = bref
    }

readOperation :: String -> (TextReader -> IO a) -> Handle -> IO a
readOperation name op h = withHOpen name h $ \tr _ _ _ _ _ _ -> case tr of
  Nothing -> illegalOp name "handle is not readable" h
  Just tr -> op tr

writeOperation :: String -> (TextWriter -> IO a) -> Handle -> IO a
writeOperation name op h = withHOpen name h $ \_ tw _ _ _ _ _ -> case tw of
  Nothing -> illegalOp name "handle not writable" h
  Just tw -> op tw

seekOperation :: String -> (Seeker -> IO a) -> Handle -> IO a
seekOperation name op h = withHOpen name h $ \_ _ _ _ s _ _ -> case s of
  Nothing -> illegalOp name "handle not seekable" h
  Just s -> op s

binReadOperation :: String -> (BinaryReader -> IO a) -> Handle -> IO a
binReadOperation name op h = withHOpen name h $ \r _ br _ _ _ _ -> case br of
  Nothing -> illegalOp name (if isJust r then "handle is in text mode" else "handle not readable") h
  Just s -> op s

binWriteOperation :: String -> (BinaryWriter -> IO a) -> Handle -> IO a
binWriteOperation name op h = withHOpen name h $ \_ w _ bw _ _ _ -> case bw of
  Nothing -> illegalOp name (if isJust w then "handle is in text mode" else "handle not writable") h
  Just s -> op s

fdOperation :: String -> (Fd -> IO a) -> Handle -> IO a
fdOperation name op h = withHOpen name h $ \_ _ _ _ _ mfd _ -> case mfd of
  Nothing -> illegalOp name "fd not present" h
  Just fd -> op fd

stdin :: Handle
stdin = unsafePerformIO $ do
  let body = HUninitialized { haUninit = UR rawStdin, haConvMode = TextMode, haBufMode = LineBuffering, haFd = Just 0 }
  newHandle "<stdin>" body

stdout :: Handle
stdout = unsafePerformIO $ do
  let body = HUninitialized { haUninit = UW rawStdout, haConvMode = TextMode, haBufMode = LineBuffering, haFd = Just 1 }
  newHandle "<stdout>" body

stderr :: Handle
stderr = unsafePerformIO $ do
  let body = HUninitialized { haUninit = UW rawStderr, haConvMode = TextMode, haBufMode = NoBuffering, haFd = Just 2 }
  newHandle "<stdout>" body

openFileIn :: ConversionMode -> FilePath -> IOMode -> IO Handle
openFileIn cmode fname mode = do
  body <- case mode of
    ReadMode -> do
      f <- fReadOpen fname
      mkFromFile f URS cmode
    WriteMode -> do
      f <- fWriteOpen fname
      mkFromFile f UWS cmode
    ReadWriteMode
      | cmode == TextMode -> illegalOpWithFname "openFile" "open text file in read-write mode" (Just fname)
      | otherwise -> do
        f <- fReadWriteOpen fname
        mkFromFile f URWS cmode
    AppendMode -> do
      f <- fAppendOpen fname
      mkFromFile f UWS TextMode
  newHandle fname body

openFile :: FilePath -> IOMode -> IO Handle
openFile = openFileIn TextMode

hClose :: Handle -> IO ()
hClose h = withHandleAlt h $ \body -> do
  case body of
    HUninitialized { haUninit = u } -> closeUninitialized u
    HOpen { haReader = reader, haWriter = writer } -> 
      let
        readerClose = do
          (TR s _) <- reader
          return $ rawClose s
        writerClose = do
          (TW s) <- writer
          return $ rawClose s
      in fromJust $ readerClose `mplus` writerClose
    HClosed {} -> return ()
    HSemiClosed { haSemiClosedReader = (TR reader _) } ->
      rawClose reader
  return ((), HClosed {})

hFileSize :: Handle -> IO Integer
hFileSize = fdOperation "hFileSize" fdFileSize

hSetFileSize :: Handle -> Integer -> IO ()
hSetFileSize h size = fdOperation "hSetFileSize" (\fd -> fdSetFileSize fd size) h

hIsEOF :: Handle -> IO Bool
hIsEOF = readOperation "hIsEOF" (\(TR s _) -> sIsEOI s)

hSetBuffering :: Handle -> BufferMode -> IO ()
hSetBuffering h mode = withHandleAlt h $ \b -> case b of
  HUninitialized {} -> return ((), b{ haBufMode = mode' })
    where
      mode' = case mode of
        BlockBuffering Nothing -> BlockBuffering (Just defaultBufferSize)
        _ -> mode
  HOpen {} -> errorUsed "hSetBuffering" h
  HSemiClosed {} -> errorSemiClosed "hSetBuffering" h
  HClosed {} -> errorClosed "hSetBuffering" h

hGetBuffering :: Handle -> IO BufferMode
hGetBuffering h = withHandle h $ \b -> case b of
  HUninitialized{ haBufMode = mode } -> return mode
  HOpen{ haBufMode = mode } -> return mode
  HSemiClosed{} -> errorSemiClosed "hSetBuffering" h
  HClosed{} -> errorClosed "hSetBuffering" h

hFlush :: Handle -> IO ()
hFlush h = withHOpen "hFlush" h $ \_ w _ _ _ _ _ -> case w of
  Nothing -> return ()
  Just (TW s) -> rawFlush s

hSeek :: Handle -> SeekMode -> Integer -> IO ()
hSeek h mode d = seekOperation "hSeek" seek h
  where
    seek (Se s) = case mode of
      AbsoluteSeek -> rawSetPos s d
      RelativeSeek -> rawRelativeSeek s d
      SeekFromEnd -> rawSeekFromEnd s d

hTell :: Handle -> IO Integer
hTell = seekOperation "hTell" (\(Se s) -> rawGetPos s)

hIsOpen :: Handle -> IO Bool
hIsOpen h = withHandle h $ \b -> case b of
  HUninitialized{} -> return True
  HOpen{} -> return True
  _ -> return False

hIsClosed :: Handle -> IO Bool
hIsClosed h = withHandle h $ \b -> case b of
  HClosed{} -> return True
  _ -> return False

hIsReadable :: Handle -> IO Bool
hIsReadable h = withHandle h $ \b -> case b of
  HUninitialized{ haUninit = u } -> return $ uIsReadable u
  HOpen{ haReader = r } -> return $ isJust r
  HSemiClosed{} -> errorSemiClosed "hIsReadable" h
  HClosed{} -> errorClosed "hIsReadable" h

uIsReadable :: Uninitialized -> Bool
uIsReadable u = case u of
  UR{} -> True
  URS{} -> True
  URWS{} -> True
  _ -> False

hIsWritable :: Handle -> IO Bool
hIsWritable h = withHandle h $ \b -> case b of
  HUninitialized{ haUninit = u } -> return $ uIsWritable u
  HOpen{ haWriter = w } -> return $ isJust w
  HSemiClosed{} -> errorSemiClosed "hIsWritable" h
  HClosed{} -> errorClosed "hIsWritable" h

uIsWritable :: Uninitialized -> Bool
uIsWritable u = case u of
  UW{} -> True
  UWS{} -> True
  URWS{} -> True
  _ -> False

hIsSeekable :: Handle -> IO Bool
hIsSeekable h = withHandle h $ \b -> case b of
  HUninitialized{ haUninit = u } -> return $ uIsSeekable u
  HOpen{ haSeeker = s } -> return $ isJust s
  HSemiClosed{} -> errorSemiClosed "hIsSeekable" h
  HClosed{} -> errorClosed "hIsSeekable" h

uIsSeekable :: Uninitialized -> Bool
uIsSeekable u = case u of
  URS{} -> True
  UWS{} -> True
  URWS{} -> True
  _ -> False

hIsTerminalDevice :: Handle -> IO Bool
hIsTerminalDevice h = withHandle h $ \b -> case b of
  HUninitialized{ haFd = mfd } -> test mfd
  HOpen{ haFd = mfd } -> test mfd
  HSemiClosed{} -> errorSemiClosed "hIsTerminalDevice" h
  HClosed{} -> errorClosed "hIsTerminalDevice" h
  where
    test Nothing = return False
    test (Just fd) = fdIsTerminal fd

hSetEcho :: Handle -> Bool -> IO ()
hSetEcho h set = fdOperation "hSetEcho" (\fd -> fdSetEcho fd set) h

hGetEcho :: Handle -> IO Bool
hGetEcho = fdOperation "hGetEcho" fdGetEcho

hShow :: Handle -> IO String
hShow h = withHandle h $ \b -> case b of
  HUninitialized{ haUninit = u, haConvMode = conv, haBufMode = buf } -> 
    format (uIsReadable u) (uIsWritable u) conv buf
  HOpen{ haReader = r, haWriter = w, haConvMode = conv, haBufMode = buf } ->
    format (isJust r) (isJust w) conv buf
  HSemiClosed{} -> return "{semi-closed}"
  HClosed{} -> return "{closed}"
  where
    format readable writable cmode bmode = do
      ty <- case (readable, writable) of
        (True, True) -> return "read-writable"
        (True, False) -> return "readable"
        (False, True) -> return "writable"
        _ -> invariantViolation "hShow" "handle is not readable nor writable"
      conv <- case cmode of
        TextMode -> return "text"
        BinaryMode -> return "binary"
        Latin1Mode -> return "latin1"
      buf <- case bmode of
        LineBuffering -> return "line"
        NoBuffering -> return "none"
        BlockBuffering (Just n) -> return $ "block (" ++ show n ++ ")"
        _ -> invariantViolation "hShow" "block size unspecified"
      return $ "{loc=" ++ haName h ++ ",type=" ++ ty ++ ",conversion=" ++ conv ++ ",buffering=" ++ buf ++ "}"

hWaitForInput :: Handle -> Int -> IO Bool
hWaitForInput h msec = withHOpen "hWaitForInput" h $ \r _ _ _ _ mfd _ -> do
  (TR s _) <- maybe (illegalOp "hWaitForInput" "handle is not readable" h) return r
  fd <- maybe (illegalOp "hWaitForInput" "fd is not present" h) return mfd
  wait s fd
  where
    wait s fd = do
      n <- sElementsInMemory s
      if n /= 0
        then return True
        else fdWaitForInput fd False {- assume non-socket fd here -} msec
        -- This is not correct because newly arrived data may not be
        -- immediately available at the user's end. However, this
        -- works in most cases. For correct implementation, clock
        -- handling will be necessary.

hGetChar :: Handle -> IO Char
hGetChar = readOperation "hGetChar" $ \(TR s _) -> rawGet s

hGetLine :: Handle -> IO String
hGetLine = readOperation "hGetLine" $ \(TR s _) -> sGetLine s

hLookAhead :: Handle -> IO Char
hLookAhead = readOperation "hLookAhead" $ \(TR _ lookahead) -> lookahead

hGetContents :: Handle -> IO String
hGetContents h = withHandleAlt h $ \b -> case b of
  HOpen{ haReader = mr, haFd = mfd } -> get mr mfd
  HUninitialized{ haUninit = u, haConvMode = cmode, haFd = mfd, haBufMode = bmode } -> do
    HOpen{ haReader = mr, haFd = mfd } <- initialize u bmode cmode mfd
    get mr mfd
  HSemiClosed{} -> errorSemiClosed "hGetContents" h
  HClosed{} -> errorClosed "hGetContents" h
  where
    get mr mfd = case mr of
      Nothing -> illegalOp "hGetContents" "handle not readable" h
      Just reader@(TR s _) -> do
        r <- rawGetContents s
        closer <- unsafeInterleaveIO $ withHandleAlt h $ \_ -> return ([], HClosed{})
        return (r ++ closer, HSemiClosed{ haSemiClosedReader = reader, haFd = mfd })

hPutChar :: Handle -> Char -> IO ()
hPutChar h c = writeOperation "hPutChar" (\(TW s) -> rawPut s c) h

hPutStr :: Handle -> String -> IO ()
hPutStr h str = writeOperation "hPutStr" (\(TW s) -> rawPuts s str) h

hPutStrLn :: Handle -> String -> IO ()
hPutStrLn h str = writeOperation "hPutStrLn" (\(TW s) -> rawPuts s str >> rawPut s '\n') h

openBinaryFile :: FilePath -> IOMode -> IO Handle
openBinaryFile = openFileIn BinaryMode

openLatin1File :: FilePath -> IOMode -> IO Handle
openLatin1File = openFileIn Latin1Mode

hSetConvMode :: String -> Handle -> Bool -> ConversionMode -> IO ()
hSetConvMode loc h use cmode = withHandleAlt h $ \b -> case b of
  HUninitialized{ haUninit = u }
    | use -> return ((), b{ haConvMode = cmode })
    | otherwise -> if uIsReadable u && uIsWritable u
      then illegalOp loc "attempt to set text mode to read-write handle" h
      else return ((), b{ haConvMode = TextMode })
  HOpen{} -> errorUsed loc h
  HSemiClosed{} -> errorSemiClosed loc h
  HClosed{} -> errorClosed loc h

hSetBinaryMode :: Handle -> Bool -> IO ()
hSetBinaryMode h binary = hSetConvMode "hSetBinaryMode" h binary BinaryMode

hSetLatin1Mode :: Handle -> Bool -> IO ()
hSetLatin1Mode h latin1 = hSetConvMode "hSetLatin1Mode" h latin1 Latin1Mode

hPutBuf :: Handle -> Ptr a -> Int -> IO ()
hPutBuf h buf size = binWriteOperation "hPutBuf" put h
  where
    put (BW s) = do
      n <- sPutBuf s (castPtr buf) size
      when (n < size) $ ioError $ mkIOError fullErrorType "System.IO.SSC.Haskell98.hPutBuf" Nothing (Just (haName h))

hGetBuf :: Handle -> Ptr a -> Int -> IO Int
hGetBuf h buf size = binReadOperation "hGetBuf" (\(BR s) -> sGetBuf s (castPtr buf) size) h

hPutBufNonBlocking :: Handle -> Ptr a -> Int -> IO Int
hPutBufNonBlocking h buf size = binWriteOperation "hPutBufNonBlocking"
  (\(BW s) -> liftM (fromMaybe 0) $ sPutBufNonBlocking s (castPtr buf) size) h

hGetBufNonBlocking :: Handle -> Ptr a -> Int -> IO Int
hGetBufNonBlocking h buf size = binReadOperation "hGetBufNonBlocking"
  (\(BR s) -> liftM (fromMaybe 0) $ sGetBufNonBlocking s (castPtr buf) size) h

openTempFile :: FilePath -> String -> IO (FilePath, Handle)
openTempFile = openLatin1TempFile

openBinaryTempFile :: FilePath -> String -> IO (FilePath, Handle)
openBinaryTempFile dir template = do
  (path, f) <- fOpenTemp dir template
  body <- mkFromFile f URWS BinaryMode
  h <- newHandle path body
  return (path, h)

openLatin1TempFile :: FilePath -> String -> IO (FilePath, Handle)
openLatin1TempFile dir template = do
  r@(_, h) <- openBinaryTempFile dir template
  hSetLatin1Mode h True
  return r
