{-# OPTIONS_GHC -fglasgow-exts #-}
{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental

   Transformer that takes memory IO port which handles encoded
   character sequence, and presents block IO interface with a
   different encoding. The encoding conversion is done using iconv.
-}
module System.IO.SSC.Iconv
  ( iIconvBlock
  , oIconvBlock
  , ioIconvBlock
  , iIconvChar
  , oIconvChar
  , ioIconvChar
  , IIconvBlock
  , OIconvBlock
  , IOIconvBlock
  ) where

import Control.Monad
import Data.Word
import Foreign.C.Error
import Foreign.C.String
import Foreign.C.Types
import Foreign.Marshal.Alloc
import Foreign.Ptr
import Foreign.Storable
import Foreign.ForeignPtr
import System.IO.SSC.Classes
import System.IO.SSC.Convert
import System.IO.SSC.Internal.Debug

foreign import ccall unsafe "iconv_block.h iconv_block_new_iconv_d" c_new_iconv_d :: CString -> CString -> IO (Ptr ())
foreign import ccall unsafe "iconv_block.h &iconv_block_delete_iconv_d" c_delete_iconv_d :: FunPtr (Ptr () -> IO ())
foreign import ccall unsafe "iconv_block.h iconv_block_convert" c_convert :: Ptr () -> Ptr CChar -> CSize -> Ptr CChar -> CSize -> IO CInt
foreign import ccall unsafe "iconv_block.h iconv_block_last_read" c_last_read :: Ptr () -> IO CInt
foreign import ccall unsafe "iconv_block.h iconv_block_last_written" c_last_written :: Ptr () -> IO CInt

type IconvD = ForeignPtr ()

type IOIconvBlock d pos base a from to = IOConversionBlock d pos base a from to
type IIconvBlock d pos base a from to = IConversionBlock d pos base a from to
type OIconvBlock d pos base a from to = OConversionBlock d pos base a from to

newIconvD :: String -> String -> IO IconvD
newIconvD from to = 
  withCString from $ \cfrom ->
  withCString to $ \cto -> do
  p <- throwErrnoIfNull "Iconv.newIconvD" $
    c_new_iconv_d cfrom cto
  newForeignPtr c_delete_iconv_d p

unPtr :: Ptr a -> a -- type inference hack
unPtr = undefined

directConvert :: (Storable a, Storable b) => IconvD -> Ptr a -> Int -> Ptr b -> Int -> IO (Int, Int, Bool)
directConvert d iptr isize optr osize
  | isize == 0 || osize == 0 = return (0, 0, True)
  | otherwise = withForeignPtr d $ \p -> do
  r <- c_convert p 
    (castPtr iptr) (fromIntegral $ isize * iesize)
    (castPtr optr) (fromIntegral $ osize * oesize)
  read <- c_last_read p
  written <- c_last_written p
  return (fromIntegral read `div` iesize, fromIntegral written `div` oesize, r /= -1)
  where
    iesize = sizeOf $ unPtr iptr
    oesize = sizeOf $ unPtr optr

convert :: (Storable a, Storable b) => IconvD -> Ptr a -> Int -> Ptr b -> Int -> IO (Int, Int)
convert d iptr isize optr osize = do
  (read, written, succeed) <- directConvert d iptr isize optr osize
  if succeed
    then return (read, written)
    else ioError $ userError "System.IO.SSC.Iconv: invalid byte sequence"

bigEndian :: IO Bool
bigEndian = alloca $ \p -> do
  poke p (1 :: Word32)
  v <- peek (castPtr p)
  return $ v == (0 :: Word8)

charEncoding :: IO String
charEncoding = do
  be <- bigEndian
  return $ if be
    then "UTF-32BE"
    else "UTF-32LE"

iconvConverter :: (Storable a, Storable b) => String -> String -> IO (Converter a b)
iconvConverter from to = do
  d <- newIconvD from to
  return $ convert d

-- | Create converter from Char to multibyte string,
-- with a handler of unrepresentable characters.
-- When a character that cannot be represented with the destination
-- character encoding is encountered, the handler is applied to the
-- character, and the result is used as an alternative
-- representation of the character. If the result is Nothing, or
-- again contains unrepresentable characters, an error is raised.
charToMbsConverter :: (Storable b) => String -> (Char -> Maybe String) -> IO (Converter Char b)
charToMbsConverter to handler = do
  from <- charEncoding
  d <- newIconvD from to
  return $ converterWithFallback (directConvert d) fallback
  where
    fallback iptr size
      | size == 0 = return (0, [])
      | otherwise = do
        c <- peek iptr
        case handler c of
          Nothing -> ioError $ userError "System.IO.SSC.Iconv: unrepresentable character"
          Just alt -> return (1, alt)

ioIconvBlock 
  :: (Storable base, Storable top)
  => String -- ^ Encoding used by the underlying port.
  -> String -- ^ Encoding to be used with the transformed port.
  -> m -- ^ Underlying port object.
  -> IO (IOIconvBlock d pos m a base top)
ioIconvBlock base top mem = do
  up <- iconvConverter base top
  dn <- iconvConverter top base
  ioConversionBlock up dn mem

iIconvBlock 
  :: (Storable base, Storable top)
  => String -- ^ Encoding used by the underlying port.
  -> String -- ^ Encoding to be used with the transformed port.
  -> m -- ^ Underlying port object.
  -> IO (IIconvBlock d pos m a base top)
iIconvBlock base top mem = do
  up <- iconvConverter base top
  iConversionBlock up mem

oIconvBlock 
  :: (Storable base, Storable top)
  => String -- ^ Encoding used by the underlying port.
  -> String -- ^ Encoding to be used with the transformed port.
  -> m -- ^ Underlying port object.
  -> IO (OIconvBlock d pos m a base top)
oIconvBlock base top mem = do
  down <- iconvConverter top base
  oConversionBlock down mem

-- | Special version of 'ioIconvBlock' that presents Haskell Char
-- interface.
ioIconvChar :: (Storable base) => String -> m -> IO (IOIconvBlock d pos m a base Char)
ioIconvChar base mem = do
  top <- charEncoding
  up <- iconvConverter base top
  dn <- charToMbsConverter base (const (Just ""))
  ioConversionBlock up dn mem

-- | Special version of 'iIconvBlock' that presents Haskell Char
-- interface.
iIconvChar :: (Storable base) => String -> m -> IO (IIconvBlock d pos m a base Char)
iIconvChar base mem = do
  enc <- charEncoding
  iIconvBlock base enc mem

-- | Special version of 'oIconvBlock' that presents Haskell Char
-- interface.
oIconvChar :: (Storable base) => String -> m -> IO (OIconvBlock d pos m a base Char)
oIconvChar base mem = do
  dn <- charToMbsConverter base (const (Just ""))
  oConversionBlock dn mem

