{-# OPTIONS_GHC -fglasgow-exts -fallow-undecidable-instances #-}
{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental

   Memory -> Stream transformer with Latin1 interpretation built in.
   For more general Memory -> Stream transformer, see
   "System.IO.SSC.Scanner".
-}
module System.IO.SSC.Latin1Scanner
  ( iLatin1Scanner
  , oLatin1Scanner
  , ioLatin1Scanner
  , Latin1Scanner
  ) where

import System.IO.SSC.Classes
import System.IO.SSC.Internal.Char
import System.IO.SSC.Internal.Ptr
import System.IO.SSC.Internal.Debug
import Control.Monad
import System.IO.Error
import System.IO.Unsafe
import Data.IORef
import Data.Typeable
import Data.Word
import Data.Char
import Foreign.Storable
import Foreign.Ptr
import Foreign.Marshal.Array

newtype Latin1Scanner d pos ba m = LMS m

iLatin1Scanner, oLatin1Scanner, ioLatin1Scanner :: m -> IO (Latin1Scanner d pos ba m)
iLatin1Scanner = latin1Scanner
oLatin1Scanner = latin1Scanner
ioLatin1Scanner = latin1Scanner

latin1Scanner :: m -> IO (Latin1Scanner d pos ba m)
latin1Scanner = return . LMS

instance (Port m) => Port (Latin1Scanner d pos ba m) where
  rawClose (LMS mem) = rawClose mem
  rawIsClosed (LMS mem) = rawIsClosed mem
  rawFlush (LMS mem) = rawFlush mem

instance (Destination dest m) => Destination dest (Latin1Scanner d pos ba m) where
  rawDest (LMS mem) = rawDest mem

instance (Lockable m) => Lockable (Latin1Scanner d pos ba m) where
  sLock (LMS mem) = sLock mem
  sTryLock (LMS mem) = sTryLock mem
  sUnsafeUnlock (LMS mem) = sUnsafeUnlock mem

instance (Discardable m) => Discardable (Latin1Scanner d pos ba m) where
  rawDiscard (LMS mem) = rawDiscard mem

instance (Seekable pos m) => Seekable pos (Latin1Scanner d pos ba m) where
  rawGetPos (LMS mem) = rawGetPos mem
  rawSetPos (LMS mem) = rawSetPos mem

instance (RelativeSeekable pos m) => RelativeSeekable pos (Latin1Scanner d pos ba m) where
  rawRelativeSeek (LMS mem) = rawRelativeSeek mem

instance (SizedInput pos m) => SizedInput pos (Latin1Scanner d pos ba m) where
  rawInputEnd (LMS mem) = rawInputEnd mem
  rawSeekFromEnd (LMS mem) = rawSeekFromEnd mem

instance (SizedOutput pos m) => SizedOutput pos (Latin1Scanner d pos ba m) where
  rawOutputEnd (LMS mem) = rawOutputEnd mem

lmsGet :: (IMemory Word8 m) => Latin1Scanner d pos ba m -> IO Char
lmsGet (LMS mem) = do
  (p0, p1) <- rawReadMem mem
  if p0 == p1
    then do
      rawEndRead mem p0
      new <- rawFillMemB mem NonBlockingPreferred
      case new of
        Nothing -> ioError $ mkIOError eofErrorType "Latin1Scanner: Get" Nothing Nothing
        Just _ -> do
          (p0', p1') <- rawReadMem mem
          get p0'
    else get p0
  where
    get ptr = do
      b <- peek ptr
      rawEndRead mem (advancePtr ptr 1)
      return $ b2c b

lmsGets :: (IMemory Word8 m) => Latin1Scanner d pos ba m -> Int -> IO [Char]
lmsGets (LMS mem) n
  | n <= 0 = return []
  | otherwise = do
    r <- loop n
    case r of
     [] -> ioError $ mkIOError eofErrorType "lmsGets" Nothing Nothing
     _ -> return r
  where
    loop 0 = return []
    loop n = do
      (p0, p1) <- rawReadMem mem
      let size = distancePtr p1 p0
      if size >= n
        then do
          let end = advancePtr p0 n
          r <- unpack p0 end
          rawEndRead mem end
          return r
        else do
          r <- unpack p0 p1
          rawEndRead mem p1
          new <- rawFillMemB mem NonBlockingPreferred
          case new of
            Nothing -> return r
            Just _ -> do
              rest <- loop (n - size)
              return (r ++ rest)

    unpack p0 p1 = seq p0 $ loop p1 []
      where
        loop p1 xs
          | p0 == p1 = return xs
          | otherwise = do
            let p1' = advancePtr p1 (-1)
            x <- peek p1'
            let c = b2c x
            c `seq` loop p1' (c:xs)

lmsGetContents :: (IMemory Word8 m, Port m) => Latin1Scanner d pos ba m -> IO [Char]
lmsGetContents lms@(LMS mem) = unsafeInterleaveIO loop
  where
    loop = do
      (p0, p1) <- rawReadMem mem
      rest <- unsafeInterleaveIO $ do
        r <- try $ rawFillMemB mem NonBlockingPreferred
        case r of
          Right (Just _) -> loop
          _ -> finish
      r <- loop1 p0 p1 rest
      rawEndRead mem p1
      return r

    loop1 p0 p1 cs
      | p0 == p1 = return cs
      | otherwise = do
        let prev = advancePtr p1 (-1)
        b <- peek prev
        loop1 p0 prev (b2c b:cs)

    finish = do
      rawClose mem
      return []

lmsGetUntil :: (IMemory Word8 m) => Latin1Scanner d pos ba m -> Word8 -> IO [Char]
lmsGetUntil (LMS mem) bend = seq bend $ do
  seen_newline <- newIORef False
  r <- getUntilLoop mem seen_newline bend
  rr <- case r of
    [] -> do
      s <- readIORef seen_newline
      if s
        then return []
        else ioError $ mkIOError eofErrorType "Latin1Scanner: rawGetUntil" Nothing Nothing
    _ -> return r
  return rr

getUntilLoop :: (IMemory Word8 m) => m -> IORef Bool -> Word8 -> IO String
getUntilLoop mem ref bend = self
  where
    self = do
      (p0, p1) <- rawReadMem mem
      let
        loop p
          | p == p1 = do
            rawEndRead mem p
            new <- rawFillMemB mem NonBlockingPreferred
            case new of
              Nothing -> return []
              Just _ -> self
          | otherwise = do
            b <- peek p
            if b == bend
              then do
                writeIORef ref True
                rawEndRead mem (advancePtr p 1)
                return []
              else do
                rest <- loop (advancePtr p 1)
                let c = b2c b
                seq c $ return (c:rest)
      p1 `seq` loop p0
        
instance (IMemory Word8 m, Port m) => IStream Char (Latin1Scanner d pos ba m) where
  rawGet  = lmsGet
  rawGets = lmsGets
  rawGetContents = lmsGetContents

instance (IMemory Word8 m, Port m) => RetrievableStream Char (Latin1Scanner d pos ba m) where
  rawGetUntil lms end = lmsGetUntil lms (c2b end)

lmsPut :: (OMemory Word8 m) => Latin1Scanner d pos ba m -> Char -> IO ()
lmsPut (LMS mem) x = do
  (p0, p1) <- rawWriteMem mem
  if p0 == p1
    then do
      rawEndWrite mem p0
      (_, stuck) <- rawFlushMemB mem NonBlockingPreferred
      if stuck
        then ioError $ mkIOError fullErrorType "Latin1Scanner: Put: Stream Stuck" Nothing Nothing
        else do
          (p0', p1') <- rawWriteMem mem
          put p0' p1'
    else put p0 p1
  where
    put p0 p1 = do
      poke p0 (c2b x)
      rawEndWrite mem (advancePtr p0 1)

{-# INLINE lmsPuts #-}
lmsPuts :: (OMemory Word8 m) => Latin1Scanner d pos ba m -> [Char] -> IO ()
lmsPuts (LMS mem) = self
  where
    self xs = do
      (p0, p1) <- {-# SCC "WRITEMEM" #-} rawWriteMem mem
      let
        loop p xs
          | p `seq` False = undefined
        loop p [] = rawEndWrite mem p
        loop p xxs@(x:xs)
          | p == p1 = do
            rawEndWrite mem p
            (_, stuck) <- rawFlushMemB mem NonBlockingPreferred
            if stuck
              then ioError $ mkIOError fullErrorType "Latin1Scanner.lmsPuts" Nothing Nothing
              else self xxs
          | otherwise = do
            poke p (c2b x)
            loop (advancePtr p 1) xs
      p1 `seq` loop p0 xs

instance (OMemory Word8 m) => OStream Char (Latin1Scanner d pos ba m) where
  rawPut = lmsPut
  rawPuts = lmsPuts

instance (IBlock a m) => IBlock a (Latin1Scanner d pos ba m) where
  rawGetBufB (LMS mem) = rawGetBufB mem

instance (OBlock a m) => OBlock a (Latin1Scanner d pos ba m) where
  rawPutBufB (LMS mem) = rawPutBufB mem

instance (IMemory a m) => IMemory a (Latin1Scanner d pos ba m) where
  rawReadMem (LMS mem) = rawReadMem mem
  rawEndRead (LMS mem) = rawEndRead mem
  rawFillMemB (LMS mem) = rawFillMemB mem

instance (OMemory a m) => OMemory a (Latin1Scanner d pos ba m) where
  rawWriteMem (LMS mem) = rawWriteMem mem
  rawEndWrite (LMS mem) = rawEndWrite mem
  rawFlushMemB (LMS mem) = rawFlushMemB mem
