{-# OPTIONS_GHC -fffi -fglasgow-exts #-}
{-# LANGUAGE ForeignFunctionInterface, MultiParamTypeClasses #-}
{- |
   License      : BSD3
   Maintainer   : Takano Akio <aljee@hyper.cx>
   Stability    : experimental

   Standard input\/output.

   For stdin, stdout, and stderr, 3 methods to access them are
   provided. One for text access respecting the current locale,
   one for binary access, and one for text access assuming latin-1
   character encoding. Different access methods do not share their
   buffers, so you cannot safely mix stdin operations with different
   methods, and you should explicitly flush the port before changing
   the access method to stdout\/stderr.

   Currently, stdout is not flushed at program exit, so you should
   flush it manually before terminating the program.
-}

module System.IO.SSC.Stdio
  ( stdin
  , stdout
  , stderr
  , binaryStdin
  , binaryStdout
  , binaryStderr
  , latin1Stdin
  , latin1Stdout
  , latin1Stderr
  , rawStdin
  , rawStdout
  , rawStderr
  , isEOI
  , interact
  , put
  , puts
  , putsLn
  , print
  , get
  , getLine
  , getContents
  , StdinType
  , StdoutType
  , StderrType
  , BinaryStdinType
  , BinaryStdoutType
  , BinaryStderrType
  , Latin1StdinType
  , Latin1StdoutType
  , Latin1StderrType
  ) where

import Data.Word
import Foreign.C.Error
import Foreign.C.Types
import System.Posix.Types(Fd)
import Prelude hiding
  ( getLine
  , print
  , getContents
  , interact
  )
import System.IO.Unsafe
import System.IO.SSC.Classes
import System.IO.SSC.Device
import System.IO.SSC.AutoLock
import System.IO.SSC.Scanner
import System.IO.SSC.Buffer
import System.IO.SSC.LocaleEncoding
import System.IO.SSC.TextMode
import System.IO.SSC.InputHook
import System.IO.SSC.AutoFlush
import System.IO.SSC.Latin1Scanner
import System.IO.SSC.Utils

rawStdin :: IDevice
rawStdin = unsafePerformIO $ do
  setBinary 0
  iDeviceFromFd 0

rawStdout :: ODevice
rawStdout = unsafePerformIO $ do
  setBinary 1
  oDeviceFromFd 1

rawStderr :: ODevice
rawStderr = unsafePerformIO $ do
  setBinary 2
  oDeviceFromFd 2

setBinary :: Fd -> IO ()
setBinary fd = 
  throwErrnoIfMinus1_ "setBinary" $
    c_set_to_binary_mode (fromIntegral fd)

foreign import ccall unsafe "stdinout.h stdinout_set_to_binary_mode"
  c_set_to_binary_mode :: CInt -> IO CInt


realStdin ::
  AutoLock Char Char Char DeviceInfo () (
    Scanner DeviceInfo () Char Char (
      IBuffer DeviceInfo () Char (
        ILocaleEncoding DeviceInfo () () (
          ITextMode DeviceInfo () () (
            IBuffer DeviceInfo () Word8 (
              InputHook DeviceInfo () () Word8 () (
                IDevice)))))))
realStdin = unsafePerformIO $ 
  return rawStdin
  >>= iInputHook (rawFlush stdout)
  >>= iBuffer
  >>= iTextMode
  >>= iNaturalEncoding 0
  >>= iBuffer
  >>= iScanner
  >>= iAutoLock

realStdout ::
  AutoLock Char Char Char DeviceInfo () (
    Scanner DeviceInfo () Char Char (
      OBuffer DeviceInfo () Char (
        OLocaleEncoding DeviceInfo () () (
          OTextMode DeviceInfo () () (
            OBuffer DeviceInfo () Word8 (
              ODevice))))))
realStdout = unsafePerformIO $
  return rawStdout
  >>= oBuffer
  >>= oTextMode
  >>= oNaturalEncoding 1
  >>= oBuffer
  >>= oScanner
  >>= oAutoLock

realStderr ::
  AutoLock Char Char Char DeviceInfo () (
    AutoFlush DeviceInfo () Char Char Char (
      Scanner DeviceInfo () Char Char (
        OBuffer DeviceInfo () Char (
          OLocaleEncoding DeviceInfo () () (
            OTextMode DeviceInfo () () (
              OBuffer DeviceInfo () Word8 (
                ODevice)))))))
realStderr = unsafePerformIO $
  return rawStderr
  >>= oBuffer
  >>= oTextMode
  >>= oNaturalEncoding 2
  >>= oBuffer
  >>= oScanner
  >>= oAutoFlush
  >>= oAutoLock

realBinaryStdin ::
  AutoLock Word8 Word8 Word8 DeviceInfo () (
    Scanner DeviceInfo () Word8 Word8 (
      IBuffer DeviceInfo () Word8 (
        InputHook DeviceInfo () () Word8 () (
          IDevice))))
realBinaryStdin = unsafePerformIO $ 
  return rawStdin
  >>= iInputHook (rawFlush binaryStdout)
  >>= iBuffer
  >>= iScanner
  >>= iAutoLock

realBinaryStdout ::
  AutoLock Word8 Word8 Word8 DeviceInfo () (
    Scanner DeviceInfo () Word8 Word8 (
      OBuffer DeviceInfo () Word8 (
        ODevice)))
realBinaryStdout = unsafePerformIO $ 
  return rawStdout
  >>= oBuffer
  >>= oScanner
  >>= oAutoLock

realBinaryStderr ::
  AutoLock Word8 Word8 Word8 DeviceInfo () (
    AutoFlush DeviceInfo () Word8 Word8 Word8 (
      Scanner DeviceInfo () Word8 Word8 (
        OBuffer DeviceInfo () Word8 (
          ODevice))))
realBinaryStderr = unsafePerformIO $ 
  return rawStderr
  >>= oBuffer
  >>= oScanner
  >>= oAutoFlush
  >>= oAutoLock

realLatin1Stdin ::
  AutoLock Char Word8 Word8 DeviceInfo () (
    Latin1Scanner DeviceInfo () Word8 (
      ITextMode DeviceInfo () () (
        IBuffer DeviceInfo () Word8 (
          InputHook DeviceInfo () () Word8 () (
            IDevice)))))
realLatin1Stdin = unsafePerformIO $ 
  return rawStdin
  >>= iInputHook (rawFlush latin1Stdout)
  >>= iBuffer
  >>= iTextMode
  >>= iLatin1Scanner
  >>= iAutoLock

realLatin1Stdout ::
  AutoLock Char Word8 Word8 DeviceInfo () (
    Latin1Scanner DeviceInfo () Word8 (
      OTextMode DeviceInfo () () (
        OBuffer DeviceInfo () Word8 (
          ODevice))))
realLatin1Stdout = unsafePerformIO $
  return rawStdout
  >>= oBuffer
  >>= oTextMode
  >>= oLatin1Scanner
  >>= oAutoLock

realLatin1Stderr ::
  AutoLock Char Word8 Word8 DeviceInfo () (
    AutoFlush DeviceInfo () Char Word8 Word8 (
      Latin1Scanner DeviceInfo () Word8 (
        OTextMode DeviceInfo () () (
          OBuffer DeviceInfo () Word8 (
            ODevice)))))
realLatin1Stderr = unsafePerformIO $
  return rawStderr
  >>= oBuffer
  >>= oTextMode
  >>= oLatin1Scanner
  >>= oAutoFlush
  >>= oAutoLock


isEOI :: IO Bool
isEOI = sIsEOI stdin

interact :: (String -> String) -> IO ()
interact f = do
  s <- getContents
  puts $ f s

put :: Char -> IO ()
put = rawPut stdout

puts :: String -> IO ()
puts = rawPuts stdout

putsLn :: String -> IO ()
putsLn = sPutsLn stdout

print :: (Show a) => a -> IO ()
print = putsLn . show

get :: IO Char
get = rawGet stdin

getLine :: IO String
getLine = sGetLine stdin

getContents :: IO String
getContents = rawGetContents stdin

stdin :: StdinType
stdin = Stdin

stdout :: StdoutType
stdout = Stdout

stderr :: StderrType
stderr = Stderr

binaryStdin :: BinaryStdinType
binaryStdin = BinaryStdin

binaryStdout :: BinaryStdoutType
binaryStdout = BinaryStdout

binaryStderr :: BinaryStderrType
binaryStderr = BinaryStderr

latin1Stdin :: Latin1StdinType
latin1Stdin = Latin1Stdin

latin1Stdout :: Latin1StdoutType
latin1Stdout = Latin1Stdout

latin1Stderr :: Latin1StderrType
latin1Stderr = Latin1Stderr

data StdinType = Stdin
data StdoutType = Stdout
data StderrType = Stderr
data BinaryStdinType = BinaryStdin
data BinaryStdoutType = BinaryStdout
data BinaryStderrType = BinaryStderr
data Latin1StdinType = Latin1Stdin
data Latin1StdoutType = Latin1Stdout
data Latin1StderrType = Latin1Stderr

instance Port StdinType where
  rawClose _ = rawClose realStdin
  rawIsClosed _ = rawIsClosed realStdin
  rawFlush _ = rawFlush realStdin

instance Destination DeviceInfo StdinType where
  rawDest _ = rawDest realStdin

instance Lockable StdinType where
  sLock _ = sLock realStdin
  sTryLock _ = sTryLock realStdin
  sUnsafeUnlock _ = sUnsafeUnlock realStdin

instance IStream Char StdinType where
  rawGet _ = rawGet realStdin
  rawGets _ = rawGets realStdin
  rawGetContents _ = rawGetContents realStdin

instance RetrievableStream Char StdinType where
  rawGetUntil _ = rawGetUntil realStdin

instance IBlock Char StdinType where
  rawGetBufB _ = rawGetBufB realStdin

instance IMemory Char StdinType where
  rawReadMem _ = rawReadMem realStdin
  rawEndRead _ = rawEndRead realStdin
  rawFillMemB _ = rawFillMemB realStdin


instance Port StdoutType where
  rawClose _ = rawClose realStdout
  rawIsClosed _ = rawIsClosed realStdout
  rawFlush _ = rawFlush realStdout

instance Destination DeviceInfo StdoutType where
  rawDest _ = rawDest realStdout

instance Lockable StdoutType where
  sLock _ = sLock realStdout
  sTryLock _ = sTryLock realStdout
  sUnsafeUnlock _ = sUnsafeUnlock realStdout

instance OStream Char StdoutType where
  rawPut _ = rawPut realStdout
  rawPuts _ = rawPuts realStdout

instance OBlock Char StdoutType where
  rawPutBufB _ = rawPutBufB realStdout

instance OMemory Char StdoutType where
  rawWriteMem _ = rawWriteMem realStdout
  rawEndWrite _ = rawEndWrite realStdout
  rawFlushMemB _ = rawFlushMemB realStdout


instance Port StderrType where
  rawClose _ = rawClose realStderr
  rawIsClosed _ = rawIsClosed realStderr
  rawFlush _ = rawFlush realStderr

instance Destination DeviceInfo StderrType where
  rawDest _ = rawDest realStderr

instance Lockable StderrType where
  sLock _ = sLock realStderr
  sTryLock _ = sTryLock realStderr
  sUnsafeUnlock _ = sUnsafeUnlock realStderr

instance OStream Char StderrType where
  rawPut _ = rawPut realStderr
  rawPuts _ = rawPuts realStderr

instance OBlock Char StderrType where
  rawPutBufB _ = rawPutBufB realStderr

instance OMemory Char StderrType where
  rawWriteMem _ = rawWriteMem realStderr
  rawEndWrite _ = rawEndWrite realStderr
  rawFlushMemB _ = rawFlushMemB realStderr


instance Port BinaryStdinType where
  rawClose _ = rawClose realBinaryStdin
  rawIsClosed _ = rawIsClosed realBinaryStdin
  rawFlush _ = rawFlush realBinaryStdin

instance Destination DeviceInfo BinaryStdinType where
  rawDest _ = rawDest realBinaryStdin

instance Lockable BinaryStdinType where
  sLock _ = sLock realBinaryStdin
  sTryLock _ = sTryLock realBinaryStdin
  sUnsafeUnlock _ = sUnsafeUnlock realBinaryStdin

instance IStream Word8 BinaryStdinType where
  rawGet _ = rawGet realBinaryStdin
  rawGets _ = rawGets realBinaryStdin
  rawGetContents _ = rawGetContents realBinaryStdin

instance RetrievableStream Word8 BinaryStdinType where
  rawGetUntil _ = rawGetUntil realBinaryStdin

instance IBlock Word8 BinaryStdinType where
  rawGetBufB _ = rawGetBufB realBinaryStdin

instance IMemory Word8 BinaryStdinType where
  rawReadMem _ = rawReadMem realBinaryStdin
  rawEndRead _ = rawEndRead realBinaryStdin
  rawFillMemB _ = rawFillMemB realBinaryStdin


instance Port BinaryStdoutType where
  rawClose _ = rawClose realBinaryStdout
  rawIsClosed _ = rawIsClosed realBinaryStdout
  rawFlush _ = rawFlush realBinaryStdout

instance Destination DeviceInfo BinaryStdoutType where
  rawDest _ = rawDest realBinaryStdout

instance Lockable BinaryStdoutType where
  sLock _ = sLock realBinaryStdout
  sTryLock _ = sTryLock realBinaryStdout
  sUnsafeUnlock _ = sUnsafeUnlock realBinaryStdout

instance OStream Word8 BinaryStdoutType where
  rawPut _ = rawPut realBinaryStdout
  rawPuts _ = rawPuts realBinaryStdout

instance OBlock Word8 BinaryStdoutType where
  rawPutBufB _ = rawPutBufB realBinaryStdout

instance OMemory Word8 BinaryStdoutType where
  rawWriteMem _ = rawWriteMem realBinaryStdout
  rawEndWrite _ = rawEndWrite realBinaryStdout
  rawFlushMemB _ = rawFlushMemB realBinaryStdout


instance Port BinaryStderrType where
  rawClose _ = rawClose realBinaryStderr
  rawIsClosed _ = rawIsClosed realBinaryStderr
  rawFlush _ = rawFlush realBinaryStderr

instance Destination DeviceInfo BinaryStderrType where
  rawDest _ = rawDest realBinaryStderr

instance Lockable BinaryStderrType where
  sLock _ = sLock realBinaryStderr
  sTryLock _ = sTryLock realBinaryStderr
  sUnsafeUnlock _ = sUnsafeUnlock realBinaryStderr

instance OStream Word8 BinaryStderrType where
  rawPut _ = rawPut realBinaryStderr
  rawPuts _ = rawPuts realBinaryStderr

instance OBlock Word8 BinaryStderrType where
  rawPutBufB _ = rawPutBufB realBinaryStderr

instance OMemory Word8 BinaryStderrType where
  rawWriteMem _ = rawWriteMem realBinaryStderr
  rawEndWrite _ = rawEndWrite realBinaryStderr
  rawFlushMemB _ = rawFlushMemB realBinaryStderr


instance Port Latin1StdinType where
  rawClose _ = rawClose realLatin1Stdin
  rawIsClosed _ = rawIsClosed realLatin1Stdin
  rawFlush _ = rawFlush realLatin1Stdin

instance Destination DeviceInfo Latin1StdinType where
  rawDest _ = rawDest realLatin1Stdin

instance Lockable Latin1StdinType where
  sLock _ = sLock realLatin1Stdin
  sTryLock _ = sTryLock realLatin1Stdin
  sUnsafeUnlock _ = sUnsafeUnlock realLatin1Stdin

instance IStream Char Latin1StdinType where
  rawGet _ = rawGet realLatin1Stdin
  rawGets _ = rawGets realLatin1Stdin
  rawGetContents _ = rawGetContents realLatin1Stdin

instance RetrievableStream Char Latin1StdinType where
  rawGetUntil _ = rawGetUntil realLatin1Stdin

instance IBlock Word8 Latin1StdinType where
  rawGetBufB _ = rawGetBufB realLatin1Stdin

instance IMemory Word8 Latin1StdinType where
  rawReadMem _ = rawReadMem realLatin1Stdin
  rawEndRead _ = rawEndRead realLatin1Stdin
  rawFillMemB _ = rawFillMemB realLatin1Stdin


instance Port Latin1StdoutType where
  rawClose _ = rawClose realLatin1Stdout
  rawIsClosed _ = rawIsClosed realLatin1Stdout
  rawFlush _ = rawFlush realLatin1Stdout

instance Destination DeviceInfo Latin1StdoutType where
  rawDest _ = rawDest realLatin1Stdout

instance Lockable Latin1StdoutType where
  sLock _ = sLock realLatin1Stdout
  sTryLock _ = sTryLock realLatin1Stdout
  sUnsafeUnlock _ = sUnsafeUnlock realLatin1Stdout

instance OStream Char Latin1StdoutType where
  rawPut _ = rawPut realLatin1Stdout
  rawPuts _ = rawPuts realLatin1Stdout

instance OBlock Word8 Latin1StdoutType where
  rawPutBufB _ = rawPutBufB realLatin1Stdout

instance OMemory Word8 Latin1StdoutType where
  rawWriteMem _ = rawWriteMem realLatin1Stdout
  rawEndWrite _ = rawEndWrite realLatin1Stdout
  rawFlushMemB _ = rawFlushMemB realLatin1Stdout


instance Port Latin1StderrType where
  rawClose _ = rawClose realLatin1Stderr
  rawIsClosed _ = rawIsClosed realLatin1Stderr
  rawFlush _ = rawFlush realLatin1Stderr

instance Destination DeviceInfo Latin1StderrType where
  rawDest _ = rawDest realLatin1Stderr

instance Lockable Latin1StderrType where
  sLock _ = sLock realLatin1Stderr
  sTryLock _ = sTryLock realLatin1Stderr
  sUnsafeUnlock _ = sUnsafeUnlock realLatin1Stderr

instance OStream Char Latin1StderrType where
  rawPut _ = rawPut realLatin1Stderr
  rawPuts _ = rawPuts realLatin1Stderr

instance OBlock Word8 Latin1StderrType where
  rawPutBufB _ = rawPutBufB realLatin1Stderr

instance OMemory Word8 Latin1StderrType where
  rawWriteMem _ = rawWriteMem realLatin1Stderr
  rawEndWrite _ = rawEndWrite realLatin1Stderr
  rawFlushMemB _ = rawFlushMemB realLatin1Stderr

