# include "fdio.h"

# include <unistd.h>
# include <stddef.h>
# include <errno.h>
# include <fcntl.h>
# include <stdio.h>
# include <sys/stat.h>
# include <stdlib.h>

# ifdef mingw32_HOST_OS
#  include <windows.h>
#  include <process.h>
# else
#  include <sys/select.h>
#  include <termios.h>
# endif



int fdio_seek_cur(void) { return SEEK_CUR; }
int fdio_seek_set(void) { return SEEK_SET; }
int fdio_seek_end(void) { return SEEK_END; }
int fdio_o_trunc(void) { return O_TRUNC; }
int fdio_o_binary(void) { return 
# ifdef O_BINARY
  O_BINARY
# else
  0
# endif
  ; }
int fdio_o_nonblock(void) { return 
# ifdef O_NONBLOCK
  O_NONBLOCK
# else
  0
# endif
  ; }
int fdio_o_noctty(void) { return
# ifdef O_NOCTTY
  O_NOCTTY
# else
  0
# endif
  ; }
int fdio_o_creat(void) { return O_CREAT; }
int fdio_o_rdonly(void) { return O_RDONLY; }
int fdio_o_wronly(void) { return O_WRONLY; }
int fdio_o_rdwr(void) { return O_RDWR; }
int fdio_o_append(void) { return O_APPEND; }
int fdio_o_excl(void) { return O_EXCL; }

struct wait_d
{
  int fd;
  int msec;
  int is_socket;
  int retval;
};

void *fdio_new_wait_d(int fd, int msec, int is_socket)
{
  struct wait_d *d = malloc(sizeof(struct wait_d));
  if(!d)
    return NULL;
  d->fd = fd;
  d->msec = msec;
  d->is_socket = is_socket;
  return d;
}

void fdio_run_wait_d(void *p)
{
  struct wait_d *d = p;
  d->retval = fdio_ready(d->fd, d->msec, d->is_socket);
}

int fdio_wait_retval(void *p)
{
  struct wait_d *d = p;
  return d->retval;
}

void fdio_delete_wait_d(void *p)
{
  free(p);
}

/* 
 * (c) The GRASP/AQUA Project, Glasgow University, 1994-2002
 *
 * hWaitForInput Runtime Support
 */

/*
 * inputReady(fd) checks to see whether input is available on the file
 * descriptor 'fd'.  Input meaning 'can I safely read at least a
 * *character* from this file object without blocking?'
 */
int
fdio_ready(int fd, int msecs, int isSock)
{
    if 
#ifndef mingw32_HOST_OS
    ( 1 ) {
#else
    ( isSock ) {
#endif
	int maxfd, ready;
	fd_set rfd;
	struct timeval tv;
        struct timeval *ptv;
	
	FD_ZERO(&rfd);
	FD_SET(fd, &rfd);
	
	/* select() will consider the descriptor set in the range of 0 to
	 * (maxfd-1) 
	 */
	maxfd = fd + 1;
        if(msecs >= 0)
        {
          tv.tv_sec  = msecs / 1000;
          tv.tv_usec = (msecs % 1000) * 1000;
          ptv = &tv;
        }
        else
          ptv = NULL;
	
	while ((ready = select(maxfd, &rfd, NULL, NULL, ptv)) < 0 ) {
	    if (errno != EINTR ) {
		return -1;
	    }
	}
	
	/* 1 => Input ready, 0 => not ready, -1 => error */
	return (ready);
    }
#ifdef mingw32_HOST_OS
    else {
	DWORD rc;
	HANDLE hFile = (HANDLE)_get_osfhandle(fd);
	DWORD avail;
        DWORD timeout = msecs >= 0 ? msecs : INFINITE;

	// WaitForMultipleObjects() works for Console input, but it
	// doesn't work for pipes (it always returns WAIT_OBJECT_0
	// even when no data is available).  There doesn't seem to be
	// an easy way to distinguish the two kinds of HANDLE, so we
	// try to detect pipe input first, and if that fails we try
	// WaitForMultipleObjects().
	//
	rc = PeekNamedPipe( hFile, NULL, 0, NULL, &avail, NULL );
	if (rc != 0) {
	    if (avail != 0) {
		return 1;
	    } else {
		return 0;
	    }
	} else {
	    rc = GetLastError();
	    if (rc == ERROR_BROKEN_PIPE) {
		return 1; // this is probably what we want
	    }
	    if (rc != ERROR_INVALID_HANDLE) {
		return -1;
	    }
	}

	rc = WaitForMultipleObjects( 1,
				     &hFile,
				     TRUE,   /* wait all */
				     timeout);
	
	/* 1 => Input ready, 0 => not ready, -1 => error */
	switch (rc) {
	case WAIT_TIMEOUT: return 0;
	case WAIT_OBJECT_0: return 1;
	default: return -1;
	}
    }
#endif
}

off_t fdio_filesize(int fd)
{
  struct stat s;
  if(fstat(fd, &s) == -1)
    return 0; /* leave errno alone */
  errno = 0;
  return s.st_size;
}

# if defined __GLASGOW_HASKELL__ && !defined mingw32_HOST_OS
int fdio_get_status_flags(int fd)
{
  return fcntl(fd, F_GETFL);
}

int fdio_set_status_flags(int fd, int flags)
{
  return fcntl(fd, F_SETFL, flags);
}
# endif

off_t fdio_lseek(int fd, int whence, off_t offset)
{
  return lseek(fd, offset, whence);
}

int fdio_setecho(int fd, int set)
{
# ifdef mingw32_HOST_OS
  HANDLE h = (HANDLE)_get_osfhandle(fd);
  DWORD mode;

  if(!GetConsoleMode(h, &mode))
  {
    errno = EINVAL; /* FIXME: Use GetLastError */
    return -1;
  }
  if(set)
    mode |= ENABLE_ECHO_INPUT | ENABLE_LINE_INPUT;
  else
    mode &= ~ENABLE_ECHO_INPUT;
  if(!SetConsoleMode(h, mode))
  {
    errno = EINVAL;
    return -1;
  }
  return 0;
# else
  struct termios t;
  if(tcgetattr(fd, &t) == -1)
    return -1;
  if(set)
    t.c_lflag |= ECHO;
  else
    t.c_lflag &= ~ECHO;
  return tcsetattr(fd, TCSANOW, &t);
# endif
}

int fdio_getecho(int fd)
{
# ifdef mingw32_HOST_OS
  HANDLE h = (HANDLE)_get_osfhandle(fd);
  DWORD mode;

  if(!GetConsoleMode(h, &mode))
  {
    errno = EINVAL; /* FIXME: Use GetLastError */
    return -1;
  }

  return (mode & ENABLE_ECHO_INPUT) != 0;
# else
  struct termios t;
  if(tcgetattr(fd, &t) == -1)
    return -1;
  return (t.c_lflag & ECHO) != 0;
# endif
}

int fdio_sync(int fd)
{
# ifdef mingw32_HOST_OS
  HANDLE h = (HANDLE)_get_osfhandle(fd);
  if(!FlushFileBuffers(h))
  {
    errno = EINVAL; /* FIXME: Use GetLastError */
    return -1;
  }
  return 0;
# else
  return fsync(fd);
# endif
}

unsigned fdio_getpid(void)
{
# ifdef mingw32_HOST_OS
  return _getpid();
# else
  return getpid();
# endif
}

int fdio_winsock_output_ready(int fd)
{
  int maxfd, ready;
  fd_set wfd;
  struct timeval tv;
  struct timeval *ptv;
  
  FD_ZERO(&wfd);
  FD_SET(fd, &wfd);
  
  /* select() will consider the descriptor set in the range of 0 to
   * (maxfd-1) 
   */
  maxfd = fd + 1;
  tv.tv_sec = tv.tv_usec = 0;
  
  while ((ready = select(maxfd, NULL, &wfd, NULL, &tv)) < 0 ) {
    if (errno != EINTR ) {
      return -1;
    }
  }
  
  /* 1 => Input ready, 0 => not ready, -1 => error */
  return (ready);
}
